/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa;

import java.net.URL;
import java.net.URLClassLoader;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.types.Schema;
import org.bluestemsoftware.specification.eoa.component.types.rt.SchemaDocument;
import org.bluestemsoftware.specification.eoa.component.types.rt.SchemaException;
import org.bluestemsoftware.specification.eoa.component.types.rt.SchemaValidationException;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.w3c.dom.Element;

/**
 * Abstract ancestor of all <code>ClassLoaders</code> defined by the EOA specification. This
 * classloader's parent classloader SHOULD be the 'java system classloader' whose constituents
 * are the system's {@link ProvidedDependency} artifacts.
 */

public abstract class AbstractClassLoader extends URLClassLoader {

    protected String id;

    /**
     * Preferred Constructor.
     * 
     * @param id
     *        identifier which SHOULD be unique within context of JVM
     * @param urls
     *        the URLs from which to load classes and resources
     */
    public AbstractClassLoader(String id, URL[] urls) {
        super(urls);
        if (id == null) {
            throw new IllegalArgumentException("id null");
        }
        this.id = id;
    }
    
    public void foo(Element myElement) throws SchemaException {
        
        // retrieve the schema by qualified name
        QName schemaName = new QName("http://mycompany.com/business/docs","MyMessages");
        Schema schema = SystemContext.getContext().getSystem().getSchema(schemaName);
        
        // retrieve its runtime provider
        SchemaDocument runtimeSchema = schema.getRuntimeProvider();
        
        // now assuming the the declaration which defines the
        // myElement instance is defined within the schema,
        // we validate it
        
        try {
            runtimeSchema.validate(myElement);
        } catch (SchemaValidationException sve) {
            // handle the error
        }
        
    }

    /**
     * Alternative constructor, i.e. in instances where the java system classloader cannot be
     * configured by <code>System</code> with 'provided dependencies', e.g. within context of
     * a test harness.
     * 
     * @param id
     *        identifier which SHOULD be unique within context of JVM
     * @param urls
     *        the URLs from which to load classes and resources
     * @param parent
     *        the parent class loader for delegation
     * 
     * @see ProvidedDependency
     */
    public AbstractClassLoader(String id, URL[] urls, ClassLoader parent) {
        super(urls, parent);
        this.id = id;
    }

    /**
     * Gets identifier which SHOULD be unique within context of JVM.
     * @return String
     */
    public String getID() {
        if (id == null) {
            throw new IllegalArgumentException("id null");
        }
        return id;
    }

}
