/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa;

import java.util.List;
import java.util.Map;

import org.bluestemsoftware.specification.eoa.component.ComponentDeployment;

/**
 * Identifies a 'component' dependency, i.e. the referencing artifact depends upon the ability
 * to retrieve and utilize one or more <code>Component</code> definitions which are defined
 * within deployment in order to implement its behavior.
 * <p>
 * Note that <code>Component</code> dependency artifacts MUST NOT be added as classpath
 * constituents. To reinforce this point, extensions to this specification SHOULD require an
 * archive extension other than 'jar'.
 * <p>
 * The rules used to distinguish a <code>ComponentDependency</code> from other dependencies
 * are an extension to the core specification.
 */
public interface ComponentDependency extends Dependency {

    /**
     * Identifies 'provided' dependencies upon which contained runtime <code>Application</code>
     * components MAY depend. Transitive dependencies, i.e. dependencies upon which a provided
     * dependency MAY depend, MUST NOT be included. Only dependencies declared explicitly
     * within referencing project are returned.
     * 
     * @return List of provided dependencies declared directly within <code>Component</code>
     *         project, or an empty list if none defined.
     */
    public List<ProvidedDependency> getProvidedDependencies();

    /**
     * Identifies 'shared' dependencies upon which contained runtime <code>Application</code>
     * components MAY depend. Transitive dependencies, i.e. dependencies upon which a shared
     * dependency MAY depend, MUST NOT be included. Only dependencies declared explicitly
     * within referencing project are returned.
     * 
     * @return List of shared dependencies declared directly within <code>Component</code>
     *         project, or an empty list if none defined.
     */
    public List<SharedDependency> getSharedDependencies();

    /**
     * Identifies 'scoped' dependencies upon which contained runtime <code>Application</code>
     * components MAY depend. Transitive dependencies, i.e. dependencies upon which a scoped
     * dependency MAY depend MUST be processed according to type specific rules and added to
     * list as indicated.
     * 
     * @return List of scoped dependencies or an empty list if none defined
     */
    public List<ScopedDependency> getScopedDependencies();

    /**
     * Identifies <code>Feature</code> definitions upon which contained runtime
     * <code>Application</code> components MAY depend.
     * 
     * @return List<FeatureDependency> or an empty list if no dependencies defined
     */
    public List<FeatureDependency> getFeatureDependencies();

    /**
     * Identifies 'module' dependencies. Transitive dependencies, i.e. dependencies upon which
     * a 'module' dependency MAY depend, MUST NOT be included. Only dependencies declared
     * explicitly within referencing project are returned.
     * 
     * @return Map of module dependencies declared directly within <code>Component</code>
     *         project each of which is mapped to a versionless ref, or an empty map if none
     *         defined.
     */
    public Map<String, ModuleDependency> getModuleDependencies();

    /**
     * Identifies <code>Component</code> dependencies upon which this component dependency
     * depends in order to implement its behavior. A compliant <code>System</code> MUST
     * process and load deployment associated with each <code>ComponentDependency</code>
     * while complying with the following restrictions:
     * <p>
     * Classpath artifacts defined by a component dependency, e.g. provided, scoped and shared
     * dependencies, MUST NOT be processed. <code>Feature</code> dependencies defined by
     * component dependency MUST NOT be processed.
     * <p>
     * See note on <code>ComponentDeployment</code> regarding loading of runtime provider
     * definitions.
     * 
     * @return List<ComponentDependency> or an empty list if none defined
     * 
     * @see ComponentDeployment#isTransitive()
     */
    public List<ComponentDependency> getComponentDependencies();

}
