/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.URI;
import java.net.URISyntaxException;

/**
 * A handle to a physical resource. The resource may or may not physically exist.
 */
public final class Resource {

    private ClassLoader classLoader;
    private String location;

    public Resource(ClassLoader classLoader, String location) {
        this.classLoader = classLoader;
        this.location = location;
    }

    /**
     * Indicates whether or not resource physically exists by attempting to open input stream
     * and then closing stream if resource exists. As an alternative, applications SHOULD
     * invoke {@link #getInputStream()} and check for <code>null</code> before actually using
     * the stream.
     * 
     * @return true if physical resource exists, false if not.
     */
    public boolean exists() {
        InputStream is = null;
        try {
            try {
                is = getInputStream();
            } catch (IOException ie) {
            }
            return is != null;
        } finally {
            if (is != null) {
                try {
                    is.close();
                } catch (IOException ignore) {
                }
            }
        }
    }

    /**
     * Retrieves input stream used to access resource.
     * <p>
     * Note that caller MUST close stream when done using it.
     * 
     * TODO: container should implement URLStreamHandlerFactory and register
     * on URL and use it to return a custom handler for HTTP that leverages
     * HTTPClient framework\feature.
     * 
     * see http://javaboutique.internet.com/tutorials/handlers2/index2.html
     * 
     * @return InputStream a new inputStream or null if resource does not exist.
     * @throws IOException
     */
    public InputStream getInputStream() throws IOException {
        URI uri = null;
        try {
            uri = new URI(location);
        } catch (URISyntaxException ue) {
            throw new IOException("Unable to parse location: " + ue);
        }
        if (!uri.isAbsolute()) {
            throw new IOException("location is not absolute.");
        }
        if (uri.getScheme().equals("classpath")) {
            String path = uri.getPath();
            while (path.startsWith("/")) {
                path = path.substring(1);
            }
            InputStream is = null;
            is = classLoader.getResourceAsStream(path);
            if (is == null) {
                throw new IOException("Resource '" + location + "' not found.");
            }
            return is;
        } else {
            return uri.toURL().openStream();
        }
    }

    /**
     * Convenience method which opens input stream, reads byte stream as characters using
     * specified encoding, closes stream and returns result text.
     * 
     * @param encoding
     *        encoding to use when converting stream from bytes to characters, e.g. 'UTF-8'.
     * @param stripIgnorableCharacters
     *        strips ISO control characters that are not whitespace and all characters that
     *        have the Unicode general category value "Cf".
     * @return String resource serialized as text.
     * @throws IOException
     *         if error occurs reading resource as text, e.g. if resource does not exist.
     */
    public String getText(String encoding, boolean stripIgnorableCharacters) throws IOException {

        InputStream inputStream = getInputStream();
        StringBuilder sb = new StringBuilder();
        try {
            Reader reader = new BufferedReader(new InputStreamReader(inputStream, encoding));
            for (int x = reader.read(); x > -1; x = reader.read()) {
                if (stripIgnorableCharacters) {
                    if (!Character.isIdentifierIgnorable(x)) {
                        sb.append((char)x);
                    }
                } else {
                    sb.append((char)x);
                }
            }
        } finally {
            try {
                if (inputStream != null)
                    inputStream.close();
            } catch (IOException ignore) {
            }
        }
        return sb.toString();

    }

}
