/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa;

import java.net.URL;

/**
 * Classloader 'shared' by all EOA deployments and by <code>System</code>. Requests for
 * classes/resources are first delegated to parent classloader, if any. If not found, this
 * classloader attempts to load class/resource. If still not found, this is an error.
 * <p>
 * Note that constituents within shared class realm SHOULD be restricted to 'specification'
 * libraries, i.e. libraries which contain only interfaces and abstract classes, e.g. javax
 * specifications, etc .... Libraries containing concrete classes, i.e. implementation classes,
 * SHOULD be scoped to deployment classloaders.
 * <p>
 * Note, to avoid linkage errors, classes within specification library MUST NOT reference
 * classes within implementation library (jvm uses the 'defining' class loader to load class
 * when resolving referenced classes, i.e. shared class loader would be asked to load the
 * referenced class -> NoClassDefFoundError.) If necessary, submit a feature request requesting
 * that authoring organization refactor any artifact that violates the above guidelines.
 * <p>
 * @see DeploymentClassLoader
 * @see SharedDependency
 */
public class SharedClassLoader extends AbstractClassLoader {

    // note we must use java 2 parent delegation when loading classes
    // even though this loader is queried last in the class loading
    // delegation hierarchy. this is b/c jvm uses defining classloader
    // to load classes referenced by classes defined by this classloader.
    // e.g. if a class within a shared artifact references java.lang.*,
    // org.bluestemsoftware.eoa.*, etc ..., this loader would be
    // queried and the request would be delegated to our parent which
    // should be java app classloader. see picture on abstract class
    // loader comment

    /**
     * Preferred Constructor
     * @param urls
     *        shared constituents
     */
    public SharedClassLoader(URL[] urls) {
        super("SharedClassLoader", urls);
    }

    /**
     * Alternative constructor, i.e. in instances where sun.misc.Launcher$AppClassLoader cannot
     * be configured by <code>System</code> with 'provided dependencies.'
     * 
     * @param parent
     * @param urls
     *        shared constituents
     */
    public SharedClassLoader(URL[] urls, ClassLoader parent) {
        super("SharedClassLoader", urls, parent);
    }

    @Override
    public String toString() {
        return id + "@" + Integer.toHexString(hashCode());
    }

}
