/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa;

import java.util.Collections;
import java.util.List;

/**
 * Identifies 'system' dependencies, i.e. represents the root of the dependency model hierarchy.
 * Provides ability to define/deploy factory deployments and component deployments and 'manage'
 * shared dependency versions and feature dependency versions.
 */
public final class SystemDependencies {

    private String organizationID;
    private String artifactID;
    private String version;
    private List<SharedDependency> sharedDependencies;
    private List<FactoryDependency> factoryDependencies;
    private List<ComponentDependency> componentDependencies;

    public SystemDependencies(String organizationID, String artifactID, String version,
            List<SharedDependency> sharedDependencies, List<FactoryDependency> factoryDependencies,
            List<ComponentDependency> componentDependencies) {
        this.organizationID = organizationID;
        this.artifactID = artifactID;
        this.version = version;
        this.sharedDependencies = sharedDependencies;
        this.factoryDependencies = factoryDependencies;
        this.componentDependencies = componentDependencies;
    }

    /**
     * Returns an identifier which identifies the organization which authored model.
     * 
     * @return String
     */
    public String getOrganizationID() {
        return organizationID;
    }

    /**
     * Returns name which uniquely identifies model within context of organization, which
     * in this case MAY serve as a proxy for system name.
     * 
     * @return String
     */
    public String getArtifactID() {
        return artifactID;
    }

    /**
     * Returns model version, i.e. revision.
     * 
     * @return String
     */
    public String getVersion() {
        return version;
    }

    /**
     * Identifies a system <code>SharedDependency</code>, i.e. a 'managed' shared
     * dependency.
     * <p>
     * Note that when configuring the shared classpath, if a shared dependency defined by a
     * <code>Deployment</code>, has the same organization id, the same artifact id but a
     * different version than a system defined shared dependency, the system defined version
     * MUST NOT be overridden with the transitive version, i.e. the 'managed' version takes
     * precedence.
     * 
     * @return List of shared dependencies or an empty list if none defined
     */
    @SuppressWarnings("unchecked")
    public final List<SharedDependency> getSharedDependencies() {
        if (sharedDependencies == null) {
            return Collections.EMPTY_LIST;
        }
        return sharedDependencies;
    }

    /**
     * Returns a system <code>SharedDependency</code> by versionless ref.
     * @param versionlessRef
     * @return shared dependency or <code>null</code> if not found.
     */
    public final SharedDependency getSharedDependency(String versionlessRef) {
        SharedDependency answer = null;
        if (versionlessRef != null && sharedDependencies != null) {
            for (SharedDependency sd : sharedDependencies) {
                if (sd.getVersionlessRef().equals(versionlessRef)) {
                    answer = sd;
                    break;
                }
            }
        }
        return answer;
    }

    /**
     * Identifies system factory dependencies. A compliant <code>System</code> MUST load all
     * listed factory definitions. A compliant <code>System</code> MUST also process and load
     * all transitive <code>Feature</code> dependencies defined by system
     * <code>FactoryDependency</code>, if any.
     * <p>
     * Note that when processing transitive feature dependencies, if a dependency has the same
     * organization id, the same artifact id but a different version than a system defined
     * feature dependency, the system defined version MUST NOT be overridden with the
     * transitive version, i.e. the 'managed' version takes precedence.
     * 
     * @return List<FactoryDependency> or an empty list if none defined
     */
    @SuppressWarnings("unchecked")
    public final List<FactoryDependency> getFactoryDependencies() {
        if (factoryDependencies == null) {
            return Collections.EMPTY_LIST;
        }
        return factoryDependencies;
    }

    /**
     * Returns a system <code>FactoryDependency</code> by versionless ref.
     * @param versionlessRef
     * @return factory dependency or <code>null</code> if not found.
     */
    public final FactoryDependency getFactoryDependency(String versionlessRef) {
        FactoryDependency answer = null;
        if (versionlessRef != null && factoryDependencies != null) {
            for (FactoryDependency ed : factoryDependencies) {
                if (ed.getVersionlessRef().equals(versionlessRef)) {
                    answer = ed;
                    break;
                }
            }
        }
        return answer;
    }

    /**
     * Identifies <code>Component</code> dependencies. A compliant <code>System</code> MUST
     * process and load each deployment associated with each <code>ComponentDependency</code>.
     * A compliant <code>System</code> MUST process <code>ComponentDependency</code>
     * artifacts transitively. 
     * <p>
     * Note that when processing transitive component and/or transitive feature dependencies,
     * and a system defined dependency exists, the managed definition MUST NOT be 'overridden'
     * by a transitive definition, i.e. the 'managed' version is fixed.
     * 
     * @return List<ComponentDependency> or an empty list if none defined
     */
    @SuppressWarnings("unchecked")
    public final List<ComponentDependency> getComponentDependencies() {
        if (componentDependencies == null) {
            return Collections.EMPTY_LIST;
        }
        return componentDependencies;
    }

}
