/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.net.URI;

import org.bluestemsoftware.specification.eoa.AbstractClassLoader;
import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage.AnyMessage;
import org.bluestemsoftware.specification.eoa.system.System.Log;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;

/**
 * Provides ability to deploy component definitions from classpath.
 */
public class ClasspathDeployment extends ComponentDeployment {

    protected AbstractClassLoader classLoader;
    protected String descriptorPath;

    /**
     * Constructors
     * 
     * @param classLoader
     *        an instance of <code>AbstractClassLoader</code> which will be used to load
     *        deployment descriptor from classpath.
     * @param descriptorPath
     *        fully qualified path of deployment descriptor.
     */
    public ClasspathDeployment(String organizationID, String artifactID, String version, String extension, AbstractClassLoader classLoader,
            String descriptorPath) {
        super(organizationID, artifactID, version, extension, new File(""), new ComponentDependencies(), false);
        if (classLoader == null) {
            throw new IllegalArgumentException("classLoader null");
        }
        this.classLoader = classLoader;
        if (classLoader == null) {
            throw new IllegalArgumentException("classLoader null");
        }
        if (descriptorPath == null) {
            throw new IllegalArgumentException("descriptorPath null");
        }
        this.descriptorPath = descriptorPath;
    }
     
   @Override
    void load(ComponentContext componentContext, Log log) throws DeploymentException {

        if (componentContext == null) {
            throw new IllegalArgumentException("componentContext == null");
        } else {
            this.deploymentContext = componentContext;
        }
        writeLock.lock();
        try {

            Document descriptorDocument = null;
            InputStream in = null;
            try {
                URI uri = new URI(descriptorPath);
                if (uri.getScheme().equals("classpath")) {
                    String path = uri.getPath();
                    while (path.startsWith("/")) {
                        path = path.substring(1);
                    }
                    in = classLoader.getResourceAsStream(path);
                    if (in == null) {
                        throw new IOException("Classpath resource '" + path + "' not found.");
                    }
                } else if (uri.getScheme() == null) {
                    String path = uri.getPath();
                    in = classLoader.getResourceAsStream(path);
                    if (in == null) {
                        throw new IOException("Classpath resource '" + path + "' not found.");
                    }
                } else {
                    throw new DeploymentException("Path '" + descriptorPath + "' invalid.");
                }
                InputSource inputSource = new InputSource(in);
                inputSource.setSystemId(descriptorPath);
                descriptorDocument = new DescriptorParser().parse(inputSource);
            } finally {
                if (in != null) {
                    try {
                        in.close();
                    } catch (IOException ie) {
                    }
                }
            }

            parseDescriptorElement(descriptorPath, descriptorDocument.getDocumentElement(), log);

        } catch (Throwable th) {
            if (!(th instanceof DeploymentException)) {
                StringBuffer sb = new StringBuffer();
                try {
                    ByteArrayOutputStream baos = new ByteArrayOutputStream();
                    th.printStackTrace(new PrintStream(baos, true, "UTF-8"));
                    sb.append(LINE_BREAK);
                    sb.append(new String(baos.toByteArray(), "UTF-8"));
                } catch (Exception exception) {
                    sb.append(th.toString());
                }
                throw new DeploymentException(sb.toString());
            } else {
                throw (DeploymentException)th;
            }
        } finally {
            writeLock.unlock();
        }

    }

    @Override
    public File getFile() {
        throw new UnsupportedOperationException("File is undefined for instances of ClasspathDeployment");
    }

    public static class BuiltInComponentsDeployment extends ClasspathDeployment {

        public static final String PATH = "classpath:///component/http.bluestemsoftware.org.specification.eoa.builtin.components.deployment.xml";

        public BuiltInComponentsDeployment(AbstractClassLoader classLoader) {
            super("org.bluestemsoftware.specification.eoa.component", "builtin", "1.0", "eoa-component", classLoader, PATH);
        }

        @Override
        void load(ComponentContext componentContext, Log log) throws DeploymentException {
            addComponent(AnyMessage.getInstance());
            super.load(componentContext, log);            
        }
        
    }

}
