/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component;

/**
 * Models a 'nested' <code>Component</code>, i.e. a <code>Component</code> (child) which
 * is nested within another <code>Component</code> (parent).
 */
public abstract class NestedComponent<R extends RootComponent, P extends Component> implements Component {

    protected R rootComponent;
    protected P parent;
    protected FragmentIdentifier fragmentIdentifier;

    public NestedComponent() {
    }

    /**
     * Retrieves parent component, i.e. component within which this component is 'nested.'
     * 
     * @return The parent component.
     */
    public P getParent() {
        return parent;
    }

    /**
     * Convenience method which returns top-level component within which this component is
     * nested.
     * <p>
     * Note that if component is nested directly beneath a top-level component, then this
     * method returns same result as {@link #getParent()}.
     * 
     * @return RootComponent
     * @throws IllegalStateException
     *         If <code>RootComponent</code> is undefined.
     */
    @SuppressWarnings("unchecked")
    public R getRootComponent() {
        if (rootComponent == null) {
            if (parent == null) {
                throw new IllegalStateException("Parent component not set.");
            }
            Component temp = parent;
            while (temp instanceof NestedComponent) {
                temp = ((NestedComponent)temp).getParent();
            }
            if (temp != null) {
                rootComponent = (R)temp;
            } else {
                throw new IllegalStateException("RootComponent is undefined.");
            }
        }
        return rootComponent;
    }

}