/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.application;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.RootComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.RuntimeProvidable;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationRT;
import org.bluestemsoftware.specification.eoa.component.engine.Engine;

/**
 * Models an <code>Application</code> component. Defines a collection of roles which are
 * 'performed' by <code>Application</code>. An <code>Application</code> is implemented by
 * an {@link Engine}.
 */
public final class Application extends RootComponent implements RuntimeProvidable {

    private Map<String, Role> roles;
    private boolean referencesEstablished;

    /**
     * Constructs an <code>Application</code> instance. Note that runtime component, i.e.
     * <code>ApplicationRT</code> is set by <code>System</code> when component is deployed.
     * 
     * @param name
     * @param roles
     */
    public Application(ComponentName name, Map<String, Role> roles) {
        if (name == null) {
            throw new IllegalArgumentException("name null");
        } else {
            this.name = name;
        }
        if (roles == null) {
            throw new IllegalArgumentException("roles null");
        } else {
            this.roles = Collections.unmodifiableMap(roles);
        }
        fragmentIdentifier = new FragmentIdentifier(name.getNamespaceURI(), Scheme.EOA, "application", name
                .getLocalPart());
        for (Role role : roles.values()) {
            role.setParent(this);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent.RuntimeProvidable#getRuntimeProvider()
     */
    public ApplicationRT getRuntimeProvider() {
        if (runtimeProvider == null) {
            throw new IllegalStateException("Runtime provider definition for Application " + name + " is undefined.");
        }
        if (!runtimeProvider.isDeployed()) {
            throw new IllegalStateException("Provider definition " + name + " not yet deployed.");
        }
        return (ApplicationRT)runtimeProvider;
    }

    /**
     * Returns a specific <code>Role</code> performed by this <code>Application</code>.
     * 
     * @param name
     *        role name
     * 
     * @return Role or <code>null</code> if undefined
     */
    public Role getRole(String name) {
        return roles.get(name);
    }

    /**
     * Returns all <code>Roles</code> performed by this <code>Application</code>.
     * 
     * @return Collection of Role objects or empty Collection if none defined.
     */
    public Collection<Role> getRoles() {
        return roles.values();
    }

    @Override
    public ComponentType getComponentType() {
        return ComponentType.APPLICATION;
    }

    @Override
    protected final void validate() throws DeploymentException {
        for (Role role : roles.values()) {
            role.validate();
        }
    }

    @Override
    protected final synchronized void establishReferences() throws DeploymentException {
        if (referencesEstablished) {
            return;
        }
        if (roles != null) {
            for (Role role : roles.values()) {
                role.establishReference();
            }
        }
        referencesEstablished = true;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(name);
        sb.append(System.getProperty("line.separator"));
        sb.append("roles:");
        sb.append(System.getProperty("line.separator"));
        sb.append(roles);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(name) + hashCode(roles);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof Application)) {
            return false;
        }

        if ((name == ((Application)obj).name || (name != null && name.equals(((Application)obj).name)))
                && (roles == ((Application)obj).roles || (roles != null && roles.equals(((Application)obj).roles)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}