/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.application.rt;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentContext;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.application.Application;
import org.bluestemsoftware.specification.eoa.ext.RootExtension.AbstractRootExtension;
import org.bluestemsoftware.specification.eoa.ext.application.ApplicationFactory;
import org.bluestemsoftware.specification.eoa.system.SystemContext;

/**
 * An abstract <code>ApplicationRT</code> ancestor which MAY be extended by type specific
 * implementations.
 */
public abstract class AbstractApplicationRT extends AbstractRootExtension implements ApplicationRT {

    protected FragmentIdentifier fragmentIdentifier;
    protected ComponentName componentName;
    protected Map<QName, ApplicationReference> applicationReferences;
    protected ComponentContext componentContext;

    private Boolean isDeployed = Boolean.FALSE;

    public AbstractApplicationRT(ApplicationFactory factory, ApplicationRT.Provider provider,
            ComponentName componentName, Map<QName, ApplicationReference> applicationReferences) {
        super(factory, provider);
        if (componentName == null) {
            throw new IllegalArgumentException("componentName null");
        } else {
            this.componentName = componentName;
        }
        fragmentIdentifier = new FragmentIdentifier(componentName.getNamespaceURI(), Scheme.EOA, "applicationRT",
                componentName.getLocalPart());
        if (applicationReferences == null) {
            throw new IllegalArgumentException("applicationReferences null");
        } else {
            this.applicationReferences = applicationReferences;
        }
        for (ApplicationReference applicationReference : applicationReferences.values()) {
            if (applicationReference instanceof AbstractApplicationReference) {
                ((AbstractApplicationReference)applicationReference).setParent(this);
            } else {
                throw new IllegalArgumentException("This impl expects instances of "
                        + AbstractApplicationReference.class.getName());
            }
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#getRuntimeProvidable()
     */
    public Application getRuntimeProvidable() {
        return SystemContext.getContext().getSystem().getApplication(componentName);
    }

    /**
     * Gets <code>ApplicationReference</code> by name.
     * 
     * @param applicationName
     * @return ApplicationReference or <code>null</code> if undefined.
     */
    public ApplicationReference getApplicationReference(QName applicationName) {
        if (applicationReferences == null) {
            return null;
        }
        return applicationReferences.get(applicationName);
    }

    /**
     * Returns all referenced 'partner' <code>Application</code> components.
     * 
     * @return Collection of <code>ApplicationReference</code> components or empty Collection
     *         if none defined.
     */
    @SuppressWarnings("unchecked")
    public Collection<ApplicationReference> getApplicationReferences() {
        if (applicationReferences == null) {
            return Collections.EMPTY_LIST;
        }
        return applicationReferences.values();
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#getName()
     */
    public ComponentName getName() {
        return componentName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#getComponentType()
     */
    public ComponentType getComponentType() {
        return ComponentType.APPLICATION;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#getComponentContext()
     */
    public ComponentContext getComponentContext() {
        return componentContext;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#deploy(org.bluestemsoftware.specification.eoa.component.ComponentContext)
     */
    public synchronized void deploy(ComponentContext componentContext) throws DeploymentException {
        if (isDeployed == null || isDeployed) {
            return; // circular refs are ok
        }
        isDeployed = null;
        this.componentContext = componentContext;
        establishReferences();
        validate();
        isDeployed = Boolean.TRUE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#isDeployed()
     */
    public synchronized boolean isDeployed() {
        return isDeployed == null ? false : isDeployed;
    }

    protected final void validate() throws DeploymentException {
        for (ApplicationReference applicationReference : applicationReferences.values()) {
            ((AbstractApplicationReference)applicationReference).validate();
        }
    }

    protected void establishReferences() throws DeploymentException {
        generateImpliedStructure();
        for (ApplicationReference applicationReference : applicationReferences.values()) {
            ((AbstractApplicationReference)applicationReference).establishReference();
        }
    }

    /**
     * Used to generate 'default' implied reference to self, i.e. if reference not explicitly
     * defined within provider definition. References to partner application MUST be explicitly
     * defined.
     * @return
     */
    protected abstract AbstractApplicationReference createMyApplicationReference() throws DeploymentException;

    protected void generateImpliedStructure() throws DeploymentException {

        // if user did not define a reference to self, we create it here.
        // references to other applications must be explicitly defined

        ApplicationReference self = getApplicationReference(componentName);
        if (self == null) {
            self = createMyApplicationReference();
            ((AbstractApplicationReference)self).setParent(this);
            applicationReferences.put(componentName, self);
        }

        applicationReferences = Collections.unmodifiableMap(applicationReferences);

    }

    public FragmentIdentifier getFragmentIdentifier() {
        return fragmentIdentifier;
    }

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("applicationName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(componentName);
        sb.append(System.getProperty("line.separator"));
        sb.append("applicationReferences: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(applicationReferences);
        return sb.toString();
    }

}