/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.binding;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.component.RootComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.RuntimeProvidable;
import org.bluestemsoftware.specification.eoa.component.binding.rt.BindingRT;
import org.bluestemsoftware.specification.eoa.component.intrface.Interface;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceFaultReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceMessageReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceOperation;

/**
 * Models a 'core' WSDL Binding component. A Binding specifies concrete message format and
 * transmission protocol details for Interface to which it is bound. The message format and
 * transmission protocol used is modeled by an <code>UnderlyingProtocol</code> component.
 */
public abstract class Binding extends RootComponent implements RuntimeProvidable {

    protected String bindingType;
    protected boolean referencesEstablished;
    protected UnderlyingProtocol underlyingProtocol;
    protected InterfaceReference interfaceReference;
    protected ElementPolicy elementPolicy;
    protected Map<QName, ? extends BindingFault> faults;
    protected Map<QName, ? extends BindingOperation> operations;
    protected Map<String, BindingAction> actions;

    /**
     * Constructs a <code>Binding</code> instance. Note that runtime provider, i.e.
     * <code>BindingRT</code> is set by <code>System</code> when component is deployed.
     * 
     * @param bindingType
     * @param name
     * @param underlyingProtocol
     * @param interfaceReference
     * @param elementPolicy
     * @param faults
     * @param operations
     * @param requiredBindingFeatures
     */
    public Binding(String bindingType, ComponentName name, UnderlyingProtocol underlyingProtocol,
            InterfaceReference interfaceReference, ElementPolicy elementPolicy,
            Map<QName, ? extends BindingFault> faults, Map<QName, ? extends BindingOperation> operations
            ) {
        if (bindingType == null) {
            throw new IllegalArgumentException("bindingType null");
        } else {
            this.bindingType = bindingType;
        }
        if (name == null) {
            throw new IllegalArgumentException("name null");
        } else {
            this.name = name;
        }
        fragmentIdentifier = new FragmentIdentifier(name.getNamespaceURI(), Scheme.EOA, "binding", name
                .getLocalPart());
        if (underlyingProtocol == null) {
            throw new IllegalArgumentException("underlyingProtocol null");
        } else {
            this.underlyingProtocol = underlyingProtocol;
            this.underlyingProtocol.setParent(this);
        }
        if (operations == null) {
            throw new IllegalArgumentException("operations null");
        } else {
            this.operations = Collections.unmodifiableMap(operations);
        }
        if (faults == null) {
            throw new IllegalArgumentException("faults null");
        } else {
            this.faults = Collections.unmodifiableMap(faults);
        }
        if (interfaceReference != null) {
            this.interfaceReference = interfaceReference;
            this.interfaceReference.setParent(this);
        }
        this.elementPolicy = elementPolicy;
        for (BindingFault fault : faults.values()) {
            fault.setParent(this);
        }
        for (BindingOperation bindingOperation : operations.values()) {
            bindingOperation.setParent(this);
        }

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent.RuntimeProvidable#getRuntimeProvider()
     */
    public BindingRT getRuntimeProvider() {
        if (runtimeProvider == null || !runtimeProvider.isDeployed()) {
            throw new IllegalStateException("Provider definition not yet deployed.");
        }
        return (BindingRT)runtimeProvider;
    }

    /**
     * Returns ref URI which identifies binding type, e.g. SOAP.
     * 
     * @return ref URI
     */
    public String getBindingType() {
        return bindingType;
    }

    /**
     * Returns underlying protocol employed by Binding.
     * 
     * @return UnderlyingProtocol
     */
    public UnderlyingProtocol getUnderlyingProtocol() {
        return underlyingProtocol;
    }

    /**
     * Returns a reference to Interface bound by this Binding, i.e. Binding is defined at
     * Interface granularity.
     * 
     * @return interfaceReference or <code>null</code> if Binding is reusable.
     * @see #isReusable()
     */
    public InterfaceReference getInterfaceReference() {
        return interfaceReference;
    }

    /**
     * Indicates whether binding is reusable, i.e. Binding specifies no BindingOperations or
     * BindingFaults, but instead uses extended Binding's default Binding rules to bind to ANY
     * Interface.
     * <p>
     * Note that reusable Bindings are NOT backwards compatible with WSDL 1.1. A Service that
     * references a reusable Binding via one of its Endpoints, cannot be described via WSDL
     * 1.1.
     * 
     * @return
     */
    public boolean isReusable() {
        return interfaceReference == null;
    }

    /**
     * Returns metadata object describing an interface operation bound by Binding.
     * <p>
     * 
     * @param operationName
     *        Qualified name of operation within bound interface. Namespace is same as inteface
     *        namespace, unless operation was inherited from an interface extended by bound
     *        interface.
     * @return BindingOperation, or null if operation not found.
     */
    public BindingOperation getOperation(QName operationName) {
        if (operations == null) {
            return null;
        }
        return operations.get(operationName);
    }

    /**
     * Gets BindingOperation objects.
     * 
     * @return Collection of operations or an empty Collection if none defined.
     */
    @SuppressWarnings("unchecked")
    public Collection<? extends BindingOperation> getOperations() {
        if (operations == null) {
            return Collections.EMPTY_LIST;
        }
        return operations.values();
    }

    /**
     * Retrieves <code>BindingAction</code> identified by a specific action URI, i.e. an
     * abstract <code>InterfaceAction</code> bound by this <code>Binding</code>.
     * 
     * @return BindingAction or <code>null</code> if undefined.
     */
    public BindingAction getBindingAction(String action) {
        if (actions == null) {
            return null;
        }
        return actions.get(action);
    }

    /**
     * Retrieves metadata describing fault available to operations defined on this Binding.
     * <p>
     * 
     * @param faultName
     *        Qualified name of fault (name of interface fault to which BindingFault is bound)
     * 
     * @return BindingFault object, or null if no fault found matching faultName.
     */
    public BindingFault getFault(QName faultName) {
        if (faults == null) {
            return null;
        }
        return faults.get(faultName);
    }

    /**
     * Gets metadata describing all faults available to operations defined on this Binding.
     * <p>
     * 
     * @return Collection of Binding faults or an empty Collection if none defined.
     */
    @SuppressWarnings("unchecked")
    public Collection<? extends BindingFault> getFaults() {
        if (faults == null) {
            return Collections.EMPTY_LIST;
        }
        return faults.values();
    }

    /**
     * Gets <code>ElementPolicy</code> attached to component.
     * 
     * @return ElementPolicy or <code>null</code> if no policy attached.
     */
    public ElementPolicy getElementPolicy() {
        return elementPolicy;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#validate()
     */
    @Override
    protected void validate() throws DeploymentException {
        if (isReusable()) {
            if (faults.size() > 0) {
                throw new DeploymentException("Binding "
                        + name
                        + " is invalid. "
                        + " A re-usable binding must not define faults.");
            }
        } else {
            Interface intrface = interfaceReference.getReferencedComponent();
            for (BindingFault bindingFault : faults.values()) {
                if (intrface.getFault(bindingFault.getFaultName()) == null) {
                    throw new DeploymentException("Binding "
                            + name
                            + " is invalid. "
                            + " Fault "
                            + bindingFault.getFaultName()
                            + " is undeclared on Interface "
                            + intrface.getName()
                            + ".");
                }
            }
        }
        if (isReusable()) {
            if (operations.size() > 0) {
                throw new DeploymentException("Binding "
                        + name
                        + " is invalid. "
                        + " A re-usable binding must not define operations.");
            }
        } else {
            Interface intrface = interfaceReference.getReferencedComponent();
            for (BindingOperation bo : operations.values()) {
                InterfaceOperation io = intrface.getOperation(bo.getName());
                if (io == null) {
                    throw new DeploymentException("Binding "
                            + name
                            + " is invalid. "
                            + " Interface operation "
                            + bo.getName()
                            + " is undefined.");
                }
                BindingMessageReference bimr = bo.getInputMessageReference();
                if (bimr != null) {
                    InterfaceMessageReference iimr = io.getInputMessageReference();
                    if (iimr == null) {
                        throw new DeploymentException("Binding "
                                + name
                                + " is invalid. "
                                + " Interface operation "
                                + bo.getName()
                                + " defines no input message ref.");
                    }
                    if (actions == null) {
                        actions = new HashMap<String,BindingAction>();
                    }
                    actions.put(iimr.getAction(), bimr);
                }
                BindingMessageReference bomr = bo.getOutputMessageReference();
                if (bomr != null) {
                    InterfaceMessageReference iomr = io.getOutputMessageReference();
                    if (iomr == null) {
                        throw new DeploymentException("Binding "
                                + name
                                + " is invalid. "
                                + " Interface operation "
                                + bo.getName()
                                + " defines no output message ref.");
                    }
                    if (actions == null) {
                        actions = new HashMap<String,BindingAction>();
                    }
                    actions.put(iomr.getAction(), bomr);
                }
                for (BindingFaultReference bfr : bo.getFaultReferences()) {
                    BindingFault bf = bfr.getFault();
                    if (bf == null) {
                        throw new DeploymentException("Binding "
                                + name
                                + " is invalid. "
                                + " Binding operation "
                                + bo.getName()
                                + " references undeclared binding fault "
                                + bfr.getFaultName()
                                + ". ");
                    }
                    QName faultName = bfr.getFault().getFaultName();
                    InterfaceFaultReference ifr = io.getFaultReference(faultName);
                    if (ifr == null) {
                        throw new DeploymentException("Binding "
                                + name
                                + " is invalid. "
                                + " Interface operation "
                                + bo.getName()
                                + " does not throw fault "
                                + faultName);
                    }
                    if (actions == null) {
                        actions = new HashMap<String,BindingAction>();
                    }
                    actions.put(ifr.getAction(), bfr);
                }
            }
        }

        // note that we do not verify that all interface operations
        // and all interface faults are bound. binding extension may
        // provide default binding rules for unbound operations

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#getComponentType()
     */
    @Override
    public ComponentType getComponentType() {
        return ComponentType.BINDING;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#establishReferences()
     */
    @Override
    protected synchronized void establishReferences() throws DeploymentException {
        if (referencesEstablished) {
            return;
        }
        if (interfaceReference != null) {
            interfaceReference.establishReference();
        }
        referencesEstablished = true;
    }

    // *********************** Identity Methods ************************************

    /**
     * Identity method which MUST be overriden by implementation. When an attempt is made to
     * import a <code>Binding</code> with the same name as a currently loaded
     * <code>Binding</code>, <code>Binding</code> components are compared for equality.
     * <p>
     * This method is used for error reporting and MUST detail the values which were used to
     * test for equalitiy.
     * 
     * @return String
     */
    public String toString() {
        return null;
    }

    /**
     * Identity method which MUST be overriden by implementation. When an attempt is made to
     * import a <code>Binding</code> with the same name as a currently loaded
     * <code>Binding</code>, <code>Binding</code> components are compared for equality.
     * 
     * @return String
     */
    public int hashCode() {
        return -1;
    }

    /**
     * Identity method which MUST be overriden by implementation. When an attempt is made to
     * import a <code>Binding</code> with the same name as a currently loaded
     * <code>Binding</code>, <code>Binding</code> components are compared for equality.
     * 
     * @return boolean
     */
    public boolean equals(Object obj) {
        return false;
    }
}