/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.engine;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentReference;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.application.Application;
import org.bluestemsoftware.specification.eoa.component.application.Role;
import org.bluestemsoftware.specification.eoa.component.intrface.Interface;
import org.bluestemsoftware.specification.eoa.component.service.Service;

/**
 * Defines an 'actor' which is responsible for performing a specific {@link Role} defined on
 * <code>Application</code> implemented by parent <code>Engine</code>. The referenced
 * <code>Service</code> fulfills the actual requirements of the <code>Role</code> as
 * defined by the <code>Interface</code> implemented by <code>Service</code>.
 */
public final class Actor extends NestedComponent<Engine, Engine> implements ComponentReference<Service> {

    private boolean referencesEstablished;
    private String roleName;
    private ComponentName serviceName;
    private Service service;
    private Role role;

    /**
     * Constructs a <code>Actor</code> component.
     * @param roleName
     *        name of role which maps to a role on <code>Application</code> implemented by
     *        reference <code>Engine</code>.
     * @param serviceName
     */
    public Actor(String roleName, ComponentName serviceName) {
        if (roleName == null) {
            throw new IllegalArgumentException("roleName null");
        } else {
            this.roleName = roleName;
        }
        if (serviceName == null) {
            throw new IllegalArgumentException("serviceName null");
        } else {
            this.serviceName = serviceName;
        }
    }

    /**
     * Gets name of role which maps to a <code>Role</code> on <code>Application</code>
     * implemented by parent <code>Engine</code>.
     * 
     * @return String name, which MUST be unique within context of parent <code>Engine</code>.
     */
    public String getRoleName() {
        return roleName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentName()
     */
    public ComponentName getReferencedComponentName() {
        return serviceName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentType()
     */
    public ComponentType getReferencedComponentType() {
        return ComponentType.SERVICE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponent()
     */
    public Service getReferencedComponent() {
        return service;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentClass()
     */
    public Class<Service> getReferencedComponentClass() {
        return Service.class;
    }

    /**
     * Returns <code>Role</code> on <code>Application</code> implemented by referenced
     * <code>Service</code>.
     * 
     * @return Role
     */
    public Role getRole() {
        return role;
    }

    final void establishReference() throws DeploymentException {
        if (referencesEstablished) {
            return;
        }
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        service = getRootComponent().getComponentContext().findReferencedComponent(this);
        if (service == null) {
            throw new DeploymentException("ComponentReference '"
                    + getFragmentIdentifier()
                    + "' is invalid. Referenced service "
                    + serviceName
                    + " is undefined.");
        }
        Application application = rootComponent.getApplicationReference().getReferencedComponent();
        role = application.getRole(roleName);
        if (role == null) {
            throw new DeploymentException("ComponentReference '"
                    + getFragmentIdentifier()
                    + "' is invalid. Referenced role "
                    + roleName
                    + " not defined on Application "
                    + application.getName()
                    + ".");
        }
        referencesEstablished = true;
    }

    final void validate() throws DeploymentException {

        // interface on referenced role MUST be interface implemented by
        // referenced Service

        Interface serviceInterface = service.getInterfaceReference().getReferencedComponent();
        Interface roleInterface = role.getReferencedComponent();
        if (!roleInterface.equals(serviceInterface)) {
            throw new DeploymentException("ServiceReference with fragmentIdentifier '"
                    + fragmentIdentifier
                    + "' is invalid. Referenced service "
                    + service.getName()
                    + " must implement interface on referenced role."
                    + " for 'myService' MUST NOT reference more than one service. "
                    + "Found: "
                    + serviceInterface.getName()
                    + ". Expected: "
                    + roleInterface.getName());
        }

    }

    public FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(parent.getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append(roleName);
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA, "actor",
                    pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(Engine parent) {
        this.parent = parent;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("roleName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(roleName);
        sb.append(System.getProperty("line.separator"));
        sb.append("serviceName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(serviceName);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(roleName) + hashCode(serviceName);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof Actor)) {
            return false;
        }

        if ((roleName == ((Actor)obj).roleName || (roleName != null && roleName.equals(((Actor)obj).roleName)))
                && (serviceName == ((Actor)obj).serviceName || (serviceName != null && serviceName
                        .equals(((Actor)obj).serviceName)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}