/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.engine.rt;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationRT;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.engine.Engine;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference.RelativeOrderBindingModuleComparator;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.RequestAction;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.ResponseAction;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ActionPolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.EffectivePolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.EndpointPolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.MessagePolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.OperationPolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.PolicyAlternative;
import org.bluestemsoftware.specification.eoa.component.policy.rt.PolicyAssertion;
import org.bluestemsoftware.specification.eoa.component.policy.rt.PolicyExpression;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ServicePolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.UnsupportedPolicyException;
import org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT;
import org.bluestemsoftware.specification.eoa.component.service.Endpoint;
import org.bluestemsoftware.specification.eoa.component.service.EndpointAction;
import org.bluestemsoftware.specification.eoa.component.service.EndpointOperation;
import org.bluestemsoftware.specification.eoa.component.service.Service;
import org.bluestemsoftware.specification.eoa.ext.feature.FeatureException;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyException;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.bluestemsoftware.specification.eoa.system.server.ApplicationFeature;
import org.bluestemsoftware.specification.eoa.system.server.BindingFeature;
import org.bluestemsoftware.specification.eoa.system.server.Feature;
import org.bluestemsoftware.specification.eoa.system.server.ModulableFeature;
import org.bluestemsoftware.specification.eoa.system.server.TransportFeature;

/**
 * References an action defined on an <code>Endpoint</code> on referenced
 * <code>Service<code>.
 */

public class EndpointActionReference implements NestedComponentRT<EngineRT, EndpointOperationReference> {

    protected FragmentIdentifier fragmentIdentifier;
    protected EndpointOperationReference parent;
    protected String action;
    protected EndpointAction endpointAction;
    protected ElementPolicy elementPolicy;
    protected MessagePolicy effectivePolicy;
    protected TransportModule transportModule;
    protected Map<String, FeatureModule> featureModules;
    protected Map<String, ApplicationModule> applicationModules;
    protected Map<Class<? extends ApplicationModule>, String> applicationModuleTypes;
    protected List<ApplicationModule> applicationModulesByRelativeOrder;
    protected Map<String, BindingModule> bindingModules;
    protected Map<Class<? extends BindingModule>, String> bindingModuleTypes;
    protected List<BindingModule> bindingModulesByRelativeOrder;

    static final String LINE_BREAK = System.getProperty("line.separator");

    public EndpointActionReference(String action, ElementPolicy elementPolicy) {
        if (action == null || action.equals("")) {
            throw new IllegalArgumentException("action null or empty string");
        } else {
            this.action = action;
        }
        this.elementPolicy = elementPolicy;
    }

    /**
     * Identifies referenced EndpointAction component.
     * 
     * @return String action
     * @see EndpointAction#getAction()
     */
    public String getAction() {
        return action;
    }

    /**
     * Gets referenced <code>EndpointAction</code> component.
     * @return
     */
    public EndpointAction getReferencedComponent() {
        return endpointAction;
    }

    /**
     * Gets <code>ElementPolicy</code> attached to this component.
     * 
     * @return elementPolicy or <code>null</code> if no policy attached.
     */
    public ElementPolicy getElementPolicy() {
        return elementPolicy;
    }

    /**
     * Returns the 'effective' policy for this Component, i.e. returns element policy.
     * 
     * @return EffectivePolicy effective policy with policy subject MESSAGE or
     *         <code>null</code> if no policy defined.
     */
    public MessagePolicy getEffectivePolicy() {
        return effectivePolicy;
    }

    /**
     * Gets all installed <code>FeatureModule</code> components.
     * 
     * @return Collection or empty collection if none defined. Note that no specific ordering
     *         is imposed.
     */
    @SuppressWarnings("unchecked")
    public Collection<FeatureModule> getFeatureModules() {
        if (featureModules == null) {
            return Collections.EMPTY_SET;
        }
        return featureModules.values();
    }

    /**
     * Returns all <code>ApplicationModule</code> components ordered according to value
     * retrieved via {@link ApplicationModule#getRelativeOrder()} using message direction
     * defined on corresponding {@link InterfaceAction}.
     * 
     * @return List of modules or an empty list if none defined.
     */
    @SuppressWarnings("unchecked")
    public List<ApplicationModule> getApplicationModules() {
        if (applicationModulesByRelativeOrder == null) {
            return Collections.EMPTY_LIST;
        }
        return applicationModulesByRelativeOrder;
    }

    /**
     * Gets a <code>ApplicationModule</code> by type ref.
     * @param ref
     * @return module or <code>null</code> if undefined.
     */
    public <T extends ApplicationModule> T getApplicationModule(Class<T> type) {
        if (applicationModules == null) {
            return null;
        }
        return type.cast(applicationModules.get(applicationModuleTypes.get(type)));
    }

    /**
     * Returns all <code>BindingModule</code> components ordered according to value retrieved
     * via {@link BindingModule#getRelativeOrder()} using message direction defined on
     * corresponding {@link InterfaceAction}.
     * 
     * @return List of modules or an empty list if none defined.
     */
    @SuppressWarnings("unchecked")
    public List<BindingModule> getBindingModules() {
        if (bindingModulesByRelativeOrder == null) {
            return Collections.EMPTY_LIST;
        }
        return bindingModulesByRelativeOrder;
    }

    /**
     * Gets a <code>BindingModule</code> by type ref.
     * @param ref
     * @return module or <code>null</code> if undefined.
     */
    public <T extends BindingModule> T getBindingModule(Class<T> type) {
        if (bindingModules == null) {
            return null;
        }
        return type.cast(bindingModules.get(bindingModuleTypes.get(type)));
    }

    /**
     * Gets <code>Transport</code> module which transports messages into/out of binding layer
     * of referenced <code>Endpoint</code>, i.e. if message direction is 'In' module MUST be
     * the first module to process message, if message direction is 'Out', module MUST be the
     * last to process the message.
     * 
     * @return TransportModule or default transport module if no module explicitly
     *         enabled/configured on this action via policy.
     * 
     * @see TransportFeature#createDefaultTransportModule(EndpointReference)
     */
    public TransportModule getTransportModule() {
        if (transportModule == null) {
            return parent.getParent().getDefaultTransportModule();
        }
        return transportModule;
    }

    /**
     * Convenience method which retrieves referenced <code>Engine</code>, i.e.
     * <code>Engine</code> component referenced by parent <code>EngineReference</code>
     * component.
     * 
     * @return
     */
    public Engine getReferencedEngine() {
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        return parent.getParent().getParent().getParent().getReferencedComponent();
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getParent()
     */
    public EndpointOperationReference getParent() {
        return parent;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getRootComponent()
     */
    public EngineRT getRootComponent() {
        return parent.getRootComponent();
    }

    void validate() throws DeploymentException {

    }

    final void establishReference() throws DeploymentException {
        if (endpointAction != null) {
            return;
        }
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        endpointAction = parent.getEndpointOperation().getEndpointAction(action);
        if (endpointAction == null) {
            throw new DeploymentException("EndpointActionReference '"
                    + getFragmentIdentifier()
                    + "' is invalid. EndpointAction '"
                    + action
                    + "', not defined on EndpointOperation '"
                    + parent.getEndpointOperation().getName()
                    + "'.");
        }
    }

    /*
     * Invoked by parent Engine component during deployment after references established.
     */
    void deriveEffectivePolicy() throws PolicyException {
        if (elementPolicy != null) {
            effectivePolicy = new MessagePolicy(elementPolicy.getMergedExpression());
        }
    }

    /*
     * Invoked by parent Engine component during deployment after effective policy derived.
     * Iterates over policy vocabulary namespaces in scope and add installs modules created by
     * instances of <code>ModulableFeature</code> which understand policy vocabulary.
     */
    void installFeatureModules() throws DeploymentException {

        Set<ModulableFeature> modulableFeatures = new HashSet<ModulableFeature>();
        for (Feature feature : SystemContext.getContext().getSystem().getServer().getFeatures()) {
            if (feature instanceof ModulableFeature) {
                modulableFeatures.add((ModulableFeature)feature);
            }
        }

        // retrieve collective set of policy vocabularies understood by all
        // features enabled on server

        Set<String> pvns = new HashSet<String>();
        for (ModulableFeature modulableFeature : modulableFeatures) {
            pvns.addAll(modulableFeature.getPolicyVocabularyNamespaces());
        }

        // retrieve public and private action policy aggregations without
        // filtering any assertions

        ActionPolicy privateActionPolicy = getPrivateActionPolicy(null);
        ActionPolicy publicActionPolicy = null;
        try {
            publicActionPolicy = getPublicActionPolicy(null);
        } catch (PolicyException pe) {
            throw new DeploymentException("Failed to install feature modules."
                    + " Error processing public policy. "
                    + pe.getMessage());
        }

        // if aggregation contains an effective expression with an assertion
        // we do not understand, throw an exception. the assumption being that
        // if assertion matches a vocabulary namespace in list, it's supported
        // unless feature throws an exception to indicate otherwise, i.e. when
        // asked to create module

        isActionPolicySupported(privateActionPolicy.getMessagePolicy(), pvns);
        isActionPolicySupported(privateActionPolicy.getOperationPolicy(), pvns);
        isActionPolicySupported(privateActionPolicy.getEndpointPolicy(), pvns);
        isActionPolicySupported(privateActionPolicy.getServicePolicy(), pvns);

        isActionPolicySupported(publicActionPolicy.getMessagePolicy(), pvns);
        isActionPolicySupported(publicActionPolicy.getOperationPolicy(), pvns);
        isActionPolicySupported(publicActionPolicy.getEndpointPolicy(), pvns);
        isActionPolicySupported(publicActionPolicy.getServicePolicy(), pvns);

        // now we iterate over modulable features and pass filtered private and public
        // action policy to create module operation. all assertions not understood
        // by feature are filtered out, which allows feature to determine whether or
        // not it supports the policy. if not, it is required to throw an exception

        Endpoint endpoint = endpointAction.getParent().getParent();
        String bindingType = endpoint.getBindingReference().getReferencedComponent().getBindingType();

        for (ModulableFeature modulableFeature : modulableFeatures) {

            Set<String> myVocabularyNamespaces = modulableFeature.getPolicyVocabularyNamespaces();

            ActionPolicy myPrivatePolicy = privateActionPolicy.filter(myVocabularyNamespaces);
            ActionPolicy myPublicPolicy = publicActionPolicy.filter(myVocabularyNamespaces);

            if (myPrivatePolicy.isEmpty() && myPublicPolicy.isEmpty()) {
                continue;
            }

            try {

                if (modulableFeature instanceof ApplicationFeature) {

                    ApplicationFeature af = (ApplicationFeature)modulableFeature;

                    if (!myPublicPolicy.isEmpty()) {
                        throw new FeatureException("Error enabling feature "
                                + af.getExtensionType()
                                + ". Public policy must not be used to"
                                + " enable/configure application features.");
                    }

                    ApplicationModule applicationModule = af.createApplicationModule(this, myPrivatePolicy);
                    if (applicationModule == null) {
                        throw new FeatureException("ApplicationFeature provider "
                                + af.getFeatureImpl()
                                + ", returned null ApplicationModule.");
                    } else {
                        addApplicationModule(applicationModule);
                        continue;
                    }

                }

                if (modulableFeature instanceof TransportFeature) {
                    TransportFeature tf = (TransportFeature)modulableFeature;
                    TransportModule transportModule = tf.createTransportModule(this, myPrivatePolicy,
                            myPublicPolicy);
                    if (transportModule == null) {
                        throw new FeatureException("TransportFeature provider "
                                + tf.getFeatureImpl()
                                + ", returned null TransportModule.");
                    } else {
                        setTransportModule(transportModule);
                        continue;
                    }
                }

                if (modulableFeature instanceof BindingFeature) {
                    BindingFeature bf = (BindingFeature)modulableFeature;
                    if (!bf.getBindingType().equals(bindingType)) {
                        continue;
                    }
                    BindingModule bindingModule = bf.createBindingModule(this, myPrivatePolicy, myPublicPolicy);
                    if (bindingModule == null) {
                        throw new FeatureException("BindingFeature provider "
                                + bf.getFeatureImpl()
                                + ", returned null BindingModule.");
                    } else {
                        addBindingModule(bindingModule);
                        continue;
                    }
                }

            } catch (FeatureException fe) {
                throw new DeploymentException("Error installing feature module on component "
                        + getFragmentIdentifier()
                        + " defined within deployment "
                        + getRootComponent().getComponentContext().getDeployment().getRef()
                        + "."
                        + fe);
            } catch (UnsupportedPolicyException pe) {
                if (pe.getMessage() != null) {
                    if (pe.getEffectivePolicy() != null) {
                        throw new DeploymentException("Error installing feature module on component "
                                + getFragmentIdentifier()
                                + " defined within deployment "
                                + getRootComponent().getComponentContext().getDeployment().getRef()
                                + ". ModulableFeature "
                                + ((Feature)modulableFeature).getExtensionType()
                                + " fails to support the following effective expression:"
                                + LINE_BREAK
                                + pe.getEffectivePolicy()
                                + LINE_BREAK
                                + pe.getMessage());
                    } else {
                        throw new DeploymentException("Error installing feature module on component "
                                + getFragmentIdentifier()
                                + " defined within deployment "
                                + getRootComponent().getComponentContext().getDeployment().getRef()
                                + ". Policy is not supported by ModulableFeature "
                                + ((Feature)modulableFeature).getExtensionType()
                                + ". "
                                + pe.getMessage());
                    }
                }
                StringBuilder sb = new StringBuilder();
                if (pe.getUnsupportedAssertions() != null) {
                    for (PolicyAssertion ua : pe.getUnsupportedAssertions()) {
                        sb.append(ua.toString());
                        sb.append(LINE_BREAK);
                    }
                }
                if (sb.length() > 0) {
                    throw new DeploymentException("Error installing feature module on component "
                            + getFragmentIdentifier()
                            + " defined within deployment "
                            + getRootComponent().getComponentContext().getDeployment().getRef()
                            + ". ModulableFeature "
                            + ((Feature)modulableFeature).getExtensionType()
                            + " fails to support the following effective expression:"
                            + LINE_BREAK
                            + pe.getEffectivePolicy()
                            + LINE_BREAK
                            + "The following assertions, at least one of which"
                            + " is required, are not supported: "
                            + LINE_BREAK
                            + sb);
                } else {
                    throw new DeploymentException("Error installing feature module on component "
                            + getFragmentIdentifier()
                            + " defined within deployment "
                            + getRootComponent().getComponentContext().getDeployment().getRef()
                            + ". ModulableFeature "
                            + ((Feature)modulableFeature).getExtensionType()
                            + " fails to support the following effective expression:"
                            + LINE_BREAK
                            + pe.getEffectivePolicy());
                }
            }

        }

        if (applicationModulesByRelativeOrder != null) {
            applicationModulesByRelativeOrder = Collections.unmodifiableList(applicationModulesByRelativeOrder);
        }
        if (bindingModulesByRelativeOrder != null) {
            bindingModulesByRelativeOrder = Collections.unmodifiableList(bindingModulesByRelativeOrder);
        }
        if (featureModules != null) {
            featureModules = Collections.unmodifiableMap(featureModules);
        }

        // the validate method has already been called by this point, so
        // perform feature module related validation now

        EndpointOperation eo = parent.getEndpointOperation();
        Binding binding = eo.getParent().getBindingReference().getReferencedComponent();

        Set<String> requiredBindingFeatures = binding.getRuntimeProvider().getRequiredBindingFeatures();
        if (requiredBindingFeatures != null) {
            for (String featureType : requiredBindingFeatures) {
                if (bindingModules == null || bindingModules.get(featureType) == null) {
                    throw new DeploymentException("Required binding feature "
                            + featureType
                            + " not enabled on component "
                            + getFragmentIdentifier()
                            + ". Required features must be explicitly enabled via ws-policy");
                }
            }
        }

        Set<String> requiredApplicationFeatures = getRootComponent().getRequiredApplicationFeatures();
        if (requiredApplicationFeatures != null) {
            for (String featureType : requiredApplicationFeatures) {
                if (applicationModules == null || applicationModules.get(featureType) == null) {
                    throw new DeploymentException("Required application feature "
                            + featureType
                            + " not enabled on component "
                            + getFragmentIdentifier()
                            + ". Required features must be explicitly enabled via ws-policy");
                }
            }
        }

        // it's o.k. if no transport module was enabled/configured via policy.
        // the default transport module will be used to provide transport behavior

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     */
    public FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(parent.getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append(action);
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA,
                    "endpointActionReference", pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(EndpointOperationReference parent) {
        this.parent = parent;
    }

    private void addApplicationModule(ApplicationModule applicationModule) throws DeploymentException {
        if (featureModules == null) {
            featureModules = new HashMap<String, FeatureModule>();
        }
        if (featureModules.containsKey(applicationModule.getExtensionType())) {
            throw new DeploymentException("A FeatureModule of type "
                    + applicationModule.getExtensionType()
                    + " is already installed on component "
                    + getFragmentIdentifier()
                    + ".");
        } else {
            applicationModule.setParent(this);
            featureModules.put(applicationModule.getExtensionType(), applicationModule);
        }
        if (applicationModules == null) {
            applicationModules = new HashMap<String, ApplicationModule>();
            applicationModuleTypes = new HashMap<Class<? extends ApplicationModule>, String>();
        }
        applicationModules.put(applicationModule.getExtensionType(), applicationModule);
        if (applicationModuleTypes.get(applicationModule.getClass()) != null) {
            throw new DeploymentException("An ApplicationModule of type "
                    + applicationModule.getClass()
                    + " is already installed on component "
                    + getFragmentIdentifier()
                    + ".");
        }
        applicationModuleTypes.put(applicationModule.getClass(), applicationModule.getExtensionType());
        if (applicationModulesByRelativeOrder == null) {
            applicationModulesByRelativeOrder = new ArrayList<ApplicationModule>();
        }
        applicationModulesByRelativeOrder.add(applicationModule);
        Collections.sort(applicationModulesByRelativeOrder, new RelativeOrderApplicationModuleComparator());
    }

    private void addBindingModule(BindingModule bindingModule) throws DeploymentException {
        if (featureModules == null) {
            featureModules = new HashMap<String, FeatureModule>();
        }
        if (featureModules.containsKey(bindingModule.getExtensionType())) {
            throw new DeploymentException("A FeatureModule of type "
                    + bindingModule.getExtensionType()
                    + " is already installed on component "
                    + getFragmentIdentifier()
                    + ".");
        } else {
            bindingModule.setParent(this);
            featureModules.put(bindingModule.getExtensionType(), bindingModule);
        }
        if (bindingModules == null) {
            bindingModules = new HashMap<String, BindingModule>();
            bindingModuleTypes = new HashMap<Class<? extends BindingModule>, String>();
        }
        bindingModules.put(bindingModule.getExtensionType(), bindingModule);
        if (bindingModuleTypes.get(bindingModule.getClass()) != null) {
            throw new DeploymentException("A BindingModule of type "
                    + bindingModule.getClass()
                    + " is already installed on component "
                    + getFragmentIdentifier()
                    + ".");
        }
        bindingModuleTypes.put(bindingModule.getClass(), bindingModule.getExtensionType());
        if (bindingModulesByRelativeOrder == null) {
            bindingModulesByRelativeOrder = new ArrayList<BindingModule>();
        }
        bindingModulesByRelativeOrder.add(bindingModule);
        Collections.sort(bindingModulesByRelativeOrder, new RelativeOrderBindingModuleComparator());
    }

    private void setTransportModule(TransportModule transportModule) throws DeploymentException {
        if (this.transportModule != null) {
            if (this.transportModule.getExtensionType().equals(transportModule.getExtensionType())) {
                return;
            }
            throw new DeploymentException("Error installing feature modules on component "
                    + getFragmentIdentifier()
                    + ". An instance of TransportModule of type "
                    + this.transportModule.getExtensionType()
                    + " already defined and yet Policy enables an additional"
                    + " module of type "
                    + transportModule.getExtensionType()
                    + ".");
        } else {
            if (featureModules == null) {
                featureModules = new HashMap<String, FeatureModule>();
            }
            featureModules.put(transportModule.getExtensionType(), transportModule);
            this.transportModule = transportModule;
            transportModule.setParent(this);
        }
    }

    private ActionPolicy getPrivateActionPolicy(Set<String> policyVocabularyNamespaces) {

        // retrieve policy with policy subject MESSAGE defined on me
        MessagePolicy msgPolicy = effectivePolicy;
        if (msgPolicy != null && policyVocabularyNamespaces != null) {
            msgPolicy = (MessagePolicy)msgPolicy.filter(policyVocabularyNamespaces);
        }

        // retrieve policy with policy subject OPERATION defined on my OperationReference
        EndpointOperationReference myOperationReference = parent;
        OperationPolicy opPolicy = myOperationReference.getEffectivePolicy();
        if (opPolicy != null && policyVocabularyNamespaces != null) {
            opPolicy = (OperationPolicy)opPolicy.filter(policyVocabularyNamespaces);
        }

        // retrieve policy with policy subject ENDPOINT defined on my EndpointReference
        EndpointReference myEndpointReference = myOperationReference.getParent();
        EndpointPolicy epPolicy = myEndpointReference.getEffectivePolicy();
        if (epPolicy != null && policyVocabularyNamespaces != null) {
            epPolicy = (EndpointPolicy)epPolicy.filter(policyVocabularyNamespaces);
        }

        // retrieve policy with policy subject SERVICE defined on my ServiceReference
        ServiceReference myServiceReference = myEndpointReference.getParent();
        ServicePolicy svcPolicy = myServiceReference.getEffectivePolicy();
        if (svcPolicy != null && policyVocabularyNamespaces != null) {
            svcPolicy = (ServicePolicy)svcPolicy.filter(policyVocabularyNamespaces);
        }

        return new ActionPolicy(true, svcPolicy, epPolicy, opPolicy, msgPolicy);

    }

    private ActionPolicy getPublicActionPolicy(Set<String> policyVocabularyNamespaces) throws PolicyException {

        // retrieve policy with policy subject MESSAGE defined on ref'd Action
        MessagePolicy msgPolicy = endpointAction.getEffectivePolicy();
        if (msgPolicy != null && policyVocabularyNamespaces != null) {
            msgPolicy = (MessagePolicy)msgPolicy.filter(policyVocabularyNamespaces);
        }

        // retrieve policy with policy subject OPERATION defined on ref'd Operation
        EndpointOperation endpointOperation = endpointAction.getParent();
        OperationPolicy opPolicy = endpointOperation.getEffectivePolicy();
        if (opPolicy != null && policyVocabularyNamespaces != null) {
            opPolicy = (OperationPolicy)opPolicy.filter(policyVocabularyNamespaces);
        }

        // retrieve policy with policy subject ENDPOINT defined on ref'd Endpoint
        Endpoint endpoint = endpointOperation.getParent();
        EndpointPolicy epPolicy = endpoint.getEffectivePolicy();
        if (epPolicy != null && policyVocabularyNamespaces != null) {
            epPolicy = (EndpointPolicy)epPolicy.filter(policyVocabularyNamespaces);
        }

        // retrieve policy with policy subject SERVICE defined on ref'd Service
        Service service = endpoint.getParent();
        ServicePolicy svcPolicy = service.getEffectivePolicy();
        if (svcPolicy != null && policyVocabularyNamespaces != null) {
            svcPolicy = (ServicePolicy)svcPolicy.filter(policyVocabularyNamespaces);
        }

        return new ActionPolicy(false, svcPolicy, epPolicy, opPolicy, msgPolicy);

    }

    private void isActionPolicySupported(EffectivePolicy effectivePolicy, Set<String> understoodVocabularies) throws DeploymentException {

        if (effectivePolicy == null) {
            return;
        }

        Set<String> opaqueAssertions = new HashSet<String>();
        PolicyExpression expression = effectivePolicy.getExpression();
        Collection<PolicyAlternative> alternatives = expression.getAlternatives();
        for (PolicyAlternative alternative : alternatives) {
            for (PolicyAssertion assertion : alternative) {
                if (understoodVocabularies.contains(assertion.getType().getNamespaceURI())) {
                    assertion.setIsSupported(true);
                } else {
                    opaqueAssertions.add(assertion.getType().getNamespaceURI());
                }
            }
        }
        if (!expression.isSupported()) {
            if (opaqueAssertions.size() > 0) {
                StringBuilder sb = new StringBuilder();
                for (String opaqueAssertion : opaqueAssertions) {
                    sb.append(opaqueAssertion);
                    sb.append(LINE_BREAK);
                }
                throw new DeploymentException("Error installing feature modules on component "
                        + getFragmentIdentifier()
                        + ". Effective expression, with policy subject "
                        + effectivePolicy.getPolicySubject()
                        + " is not supported:"
                        + LINE_BREAK
                        + expression.toString()
                        + "The following vocabularies are not understood, at least one of which"
                        + " maps to a required assertion: "
                        + LINE_BREAK
                        + sb.toString());
            } else {
                throw new DeploymentException("Error installing feature modules on component "
                        + getFragmentIdentifier()
                        + ". Effective expression, with policy subject "
                        + effectivePolicy.getPolicySubject()
                        + " is not supported:"
                        + LINE_BREAK
                        + expression.toString());
            }
        }

    }

    static class RelativeOrderApplicationModuleComparator implements Comparator<ApplicationModule> {
        public int compare(ApplicationModule o1, ApplicationModule o2) {
            return o1.getRelativeOrder() - o2.getRelativeOrder();
        }
    }

    /**
     * References a 'request' action, i.e. an action whose direction is 'In'.
     */
    public static class RequestActionReference extends EndpointActionReference {

        public RequestActionReference(String action, ElementPolicy elementPolicy) {
            super(action, elementPolicy);
        }

        /**
         * Convenience method used to create a runtime instance of referenced
         * {@link InterfaceAction}.
         * 
         * @return actionContext
         * 
         * @see RequestAction#createAction(ApplicationRT)
         */
        public ActionContext createAction() {

            // note that we retrieve application implemented by parent engine, i.e.
            // the application that will be working with the actual message instance,
            // e.g. if binding is delivering a request to my service, application is
            // the receiving node. if we're sending a request to a partner node, the
            // application retrieved here represents the sending node

            InterfaceAction ia = endpointAction.getInterfaceAction();
            ApplicationRT application = getRootComponent().getApplication();
            return ((RequestAction)ia).createAction(application);

        }

    }

    /**
     * References a 'response' action, i.e. an action whose direction is 'Out'.
     */
    public static class ResponseActionReference extends EndpointActionReference {

        public ResponseActionReference(String action, ElementPolicy elementPolicy) {
            super(action, elementPolicy);
        }

        /**
         * Convenience method used to create a runtime instance of referenced
         * {@link InterfaceAction}.
         * 
         * @return actionContext
         * 
         * @see ResponseAction#createAction(ApplicationRT, String)
         */
        public ActionContext createAction(String relatesTo) {

            // note that we retrieve application implemented by parent engine, i.e.
            // the application that will be working with the actual message instance,
            // e.g. if binding is delivering a response to my service, application is
            // the receiving node. if we're sending a response to a partner node, the
            // application retrieved here represents the sending node

            InterfaceAction ia = endpointAction.getInterfaceAction();
            ApplicationRT application = getRootComponent().getApplication();
            return ((ResponseAction)ia).createAction(application, relatesTo);

        }

    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("action: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(action);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(action);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof EndpointActionReference)) {
            return false;
        }

        if ((action == ((EndpointActionReference)obj).action || (action != null && action
                .equals(((EndpointActionReference)obj).action)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}