/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.engine.rt;

import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentReference;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationRT;
import org.bluestemsoftware.specification.eoa.component.application.rt.RoleReference;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.binding.rt.BindingRT;
import org.bluestemsoftware.specification.eoa.component.engine.Engine;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference.RelativeOrderApplicationModuleComparator;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.BusinessFault;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.FaultContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.SystemFault;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext.MessageModules;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.EndpointPolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.PolicyExpression;
import org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT;
import org.bluestemsoftware.specification.eoa.component.service.Endpoint;
import org.bluestemsoftware.specification.eoa.ext.feature.FeatureException;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.WSAMessageModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.rt.TransportProtocolRT;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyException;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.bluestemsoftware.specification.eoa.system.server.ApplicationFeature;
import org.bluestemsoftware.specification.eoa.system.server.BindingFeature;
import org.bluestemsoftware.specification.eoa.system.server.Feature;
import org.bluestemsoftware.specification.eoa.system.server.Server;
import org.bluestemsoftware.specification.eoa.system.server.TransportFeature;

/**
 * References a specific Endpoint on a Service referenced by parent ServiceReference component.
 */

public class EndpointReference implements NestedComponentRT<EngineRT, ServiceReference> {

    private FragmentIdentifier fragmentIdentifier;
    private ServiceReference parent;
    private String endpointName;
    private Endpoint endpoint;
    private EndpointPolicy effectivePolicy;
    private ElementPolicy elementPolicy;
    private Map<QName, EndpointOperationReference> operationReferences;
    private Map<String, EndpointActionReference> endpointActionReferences;
    private List<BindingModule> defaultBindingModules;
    private List<ApplicationModule> defaultApplicationModules;
    private TransportModule defaultTransportModule;
    private Set<FeatureModule> defaultFeatureModules;
    private boolean isAutoGenerated;

    /**
     * Construct an <code>EndpointReference</code> component.
     * @param endpointName
     *        Identifies name of referenced Endpoint component.
     * @param elementPolicy
     *        element policy attached to this component, or <code>null</code> if no policy
     *        attached
     * @param operationReferences
     *        EndpointOperationReference objects owned by this object or an empty Map if none
     *        defined.
     */
    public EndpointReference(String endpointName, ElementPolicy elementPolicy,
            Map<QName, EndpointOperationReference> operationReferences) {

        if (endpointName == null || endpointName.equals("")) {
            throw new IllegalArgumentException("endpointName null or empty string");
        } else {
            this.endpointName = endpointName;
        }
        if (operationReferences == null) {
            throw new IllegalArgumentException("operationReferences null");
        } else {
            this.operationReferences = operationReferences;
            for (EndpointOperationReference operationReference : operationReferences.values()) {
                operationReference.setParent(this);
            }
        }
        this.elementPolicy = elementPolicy;
        this.defaultApplicationModules = new ArrayList<ApplicationModule>();
        this.defaultBindingModules = new ArrayList<BindingModule>();
        this.defaultFeatureModules = new HashSet<FeatureModule>();
    }

    /**
     * Identifies name of referenced <code>Endpoint</code> component.
     * 
     * @return String name of referenced <code>Endpoint</code>.
     */
    public final String getEndpointName() {
        return endpointName;
    }

    /**
     * Method which MAY be used by code generators to generate address path for a referenced
     * <code>Endpoint</code> component defined on a locally hosted service. The address path
     * is formatted as follows:
     * <p>
     * '/eoa/ws/${engine.qname}/${service.qname}/${endpoint.name}/'
     * <p>
     * The following values are substituted in place of the indicated variable:
     * <p>
     * ${engine.qname}: Qualifed name of engine which 'hosts' service upon which endpoint is
     * defined. Qualified name of engine is formatted as follows (note that if local name of
     * engine is unique within context of system, the namespace declaration is optional, but is
     * nevertheless included within result for the sake of completeness):
     * <p>
     * [xmlns(ns=http://foo)ns:]MyEngine
     * <p>
     * ${service.qname}: Qualified name of referenced service component which is formatted as
     * follows (note that if local name of referenced service is unique within context of
     * engine which hosts the service, the namespace declaration is optional, but is
     * nevertheless included within result for the sake of completeness):
     * <p>
     * [xmlns(ns=http://foo)ns:]MyService
     * <p>
     * ${endpoint.name}: name of referenced <code>Endpoint</code> component followed by a
     * trailing slash to avoid redirects by target http server, i.e. in accordance with Servlet
     * path specification.
     * <p>
     * Example path:
     * <p>
     * 'eoa/ws/xmlns(ns=http://foo)ns:MyEngine/xmlns(ns=http://foo)ns:MyService/MyEndpoint/'
     * 
     * @return path which MUST be unqiue within context of <code>System</code>.
     * 
     * @see Endpoint#getAddress()
     */
    public static String generateAddressPath(EndpointReference epr) throws URISyntaxException {

        ServiceReference serviceReference = epr.getParent();

        if (!serviceReference.isMyService()) {
            throw new IllegalArgumentException("Expected a 'local' service reference.");
        }

        EngineReference engineReference = serviceReference.getParent();
        QName engineName = engineReference.getParent().getName();
        String en = "xmlns(ns=" + engineName.getNamespaceURI() + ")" + "ns:" + engineName.getLocalPart();

        QName serviceName = serviceReference.getReferencedComponentName();
        String sn = "xmlns(ns=" + serviceName.getNamespaceURI() + ")" + "ns:" + serviceName.getLocalPart();

        String path = null;
        path = "/eoa/ws/" + en + "/";
        path = path + sn + "/";
        path = path + epr.getEndpointName() + "/";
        return path;

    }

    /**
     * Method which MAY be used by code generators to generate wsa:ReplyTo address path for a
     * referenced <code>Endpoint</code> component, i.e. used by referenced service to return
     * asynchronous replies. The address path is formatted as follows:
     * <p>
     * '/eoa/ws/${engine.qname}/${engine.ref.qname}/${service.ref.qname}/${ep.name}/'
     * <p>
     * The following values are substituted in place of the indicated variable:
     * <p>
     * ${engine.qname}: Qualifed name of endpoint reference component's parent engine. The
     * Qualified name is formatted as follows (note that if local name of engine is unique
     * within context of system, the namespace declaration is optional, but is nevertheless
     * included within result for the sake of completeness):
     * <p>
     * [xmlns(ns=http://foo)ns:]MyEngine
     * <p>
     * ${engine.ref.qname}: qualified name of referenced partner engine (note that if local
     * name of partner engine is unique within context of endpoint reference component's parent
     * engine, the namespace declaration is optional, but is nevertheless included within
     * result for the sake of completeness) which is formatted as follows:
     * <p>
     * [xmlns(ns=http://foo)ns:]PartnerEngine
     * <p>
     * ${service.ref.qname}: qualifed name of referenced service component (note that if local
     * name of referenced service is unique within context of parent
     * <code>EngineReference</code> component, the namespace declaration is optional, but is
     * nevertheless included within result for the sake of completeness) which is formatted as
     * follows:
     * <p>
     * [xmlns(ns=http://foo)ns:]PartnerService
     * <p>
     * ${ep.name}: name of referenced <code>Endpoint</code> component followed by a trailing
     * slash to avoid redirects by target http server, i.e. in accordance with Servlet path
     * specification.
     * <p>
     * Example path:
     * <p>
     * 'eoa/ws/xmlns(ns=http://foo)ns:MyEngine/xmlns(ns=http://foo)ns:PartnerEngine/xmlns(ns=http://foo)ns:PartnerService/PartnerEndpoint/'
     * 
     * @return path which MUST be unqiue within context of <code>System</code>.
     */
    public static String generateReplyToAddressPath(EndpointReference epr) throws URISyntaxException {

        ServiceReference serviceReference = epr.getParent();

        QName myEngineName = epr.getRootComponent().getName();
        String men = "xmlns(ns=" + myEngineName.getNamespaceURI() + ")" + "ns:" + myEngineName.getLocalPart();

        EngineReference engineReference = serviceReference.getParent();
        QName pEngineName = engineReference.getReferencedComponentName();
        String pen = "xmlns(ns=" + pEngineName.getNamespaceURI() + ")" + "ns:" + pEngineName.getLocalPart();

        QName pServiceName = serviceReference.getReferencedComponentName();
        String psn = "xmlns(ns=" + pServiceName.getNamespaceURI() + ")" + "ns:" + pServiceName.getLocalPart();

        String path = null;
        path = "/eoa/ws/" + men + "/";
        path = path + pen + "/";
        path = path + psn + "/";
        path = path + epr.getEndpointName() + "/";
        return path;
    }

    /**
     * Gets referenced <code>Endpoint</code> component.
     * 
     * @return Endpoint referenced <code>Endpoint</code>.
     */
    public final Endpoint getEndpoint() {
        return endpoint;
    }

    /**
     * Convenience method which retrieves referenced <code>Engine</code>, i.e.
     * <code>Engine</code> component referenced by parent <code>EngineReference</code>
     * component.
     * 
     * @return
     */
    public final Engine getReferencedEngine() {
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        return parent.getParent().getReferencedComponent();
    }

    /**
     * Gets <code>ElementPolicy</code> attached to this component.
     * 
     * @return elementPolicy or <code>null</code> if no policy attached
     */
    public final ElementPolicy getElementPolicy() {
        return elementPolicy;
    }

    /**
     * Returns the 'effective' policy for this Component. Returns the merged result of element
     * Policy attached to this component and element policy attached to corresponding
     * <code>RoleReference</code> component retrieved from parent
     * <code>ServiceReference</code>.
     * 
     * @return EffectivePolicy effective policy with policy subject ENDPOINT or
     *         <code>null</code> if no policy defined.
     * @see ServiceReference#getCorrespondingApplicationReference()
     */
    public final EndpointPolicy getEffectivePolicy() {
        return effectivePolicy;
    }

    /**
     * Gets a specific EndpointOperationReference component.
     * 
     * @param operationName
     *        qualified name of operation
     * @return EndpointOperationReference
     */
    public final EndpointOperationReference getOperationReference(QName operationName) {
        if (operationReferences == null) {
            return null;
        }
        return operationReferences.get(operationName);
    }

    /**
     * Gets all EndpointOperationReference objects owned by this object.
     * 
     * @return Collection of EndpointOperationReference objects
     */
    @SuppressWarnings("unchecked")
    public final Collection<EndpointOperationReference> getOperationReferences() {
        if (operationReferences == null) {
            return Collections.EMPTY_LIST;
        }
        return Collections.unmodifiableCollection(operationReferences.values());
    }

    /**
     * Convenience method for retrieving an <code>EndpointActionReference</code> component
     * nested within this component, i.e. as opposed to iterating over all nested operations.
     * 
     * @param action
     * @return EndpointActionReference or <code>null</code> if not found
     */
    public final EndpointActionReference getEndpointActionReference(String action) {
        if (endpointActionReferences == null) {
            return null;
        }
        return endpointActionReferences.get(action);
    }

    /**
     * Convenience method for retrieving all <code>EndpointActionReference</code> component
     * nested within this component, i.e. as opposed to iterating over all nested operations.
     * 
     * @param action
     * @return Collection of EndpointActionReference components or an empty collection if none
     *         defined.
     */
    @SuppressWarnings("unchecked")
    public final Collection<EndpointActionReference> getEndpointActionReferences() {
        if (endpointActionReferences == null) {
            return Collections.EMPTY_SET;
        } else {
            return endpointActionReferences.values();
        }
    }

    /**
     * Returns <code>BindingModules</code> which MUST be enabled by default when
     * <code>Binding</code> processes undeclared action {@link SystemFault#ACTION}, i.e.
     * where implied message direction is 'Out.'
     * <p>
     * Note that if default module behavior is inappropriate, user MAY override by declaring
     * <code>SystemFault</code> as thrown and configuring via policy just as with any other
     * action.
     * 
     * @return List of binding modules listed in execution order or an empty list if no
     *         <code>BindingFeatures</code> defined as required by <code>Binding</code>.
     * 
     * @see BindingFeature#createDefaultBindingModule(String, boolean)
     * @see Binding#getRequiredBindingFeatures()
     */
    public final List<BindingModule> getDefaultBindingModules() {
        return defaultBindingModules;
    }

    /**
     * Returns <code>ApplicationModules</code> which are enabled by default when
     * <code>Engine</code> processes undeclared action {@link SystemFault#ACTION}, i.e.
     * where implied message direction is 'Out'.
     * <p>
     * Note that if default module behavior is inappropriate, user MAY override by declaring
     * <code>SystemFault</code> as thrown and configuring via policy just as with any other
     * action.
     * 
     * @return List of application modules listed in execution order or an empty list if no
     *         <code>ApplicationFeatures</code> defined as required by <code>Engine</code>.
     * 
     * @see ApplicationFeature#createSpringApplicationModule(boolean)
     * @see Engine#getRequiredApplicationFeatures()
     */
    public final List<ApplicationModule> getSpringApplicationModules() {
        return defaultApplicationModules;
    }

    /**
     * Returns default <code>TransportModule</code> which MUST be used by
     * <code>Binding</code> to process undeclared action {@link SystemFault#ACTION}, i.e.
     * where implied message direction is 'Out.'
     * <p>
     * Note that if default module behavior is inappropriate, user MAY override by declaring
     * <code>SystemFault</code> as thrown and configuring via policy just as with any other
     * action.
     * 
     * @return TransportModule which employs transport used by <code>UnderlyingTransport</code>
     *         on <code>Binding</code> employed by referenced <code>Endpoint</code>.
     * 
     * @see ApplicationFeature#createSpringApplicationModule(boolean)
     */
    public final TransportModule getDefaultTransportModule() {
        return defaultTransportModule;
    }

    public final Collection<FeatureModule> getDefaultFeatureModules() {
        return defaultFeatureModules;
    }

    /**
     * Receives an action from referenced <code>Endpoint</code>. If referenced Service upon
     * which action is defined is 'my' Service, action represents a request, i.e. direction is
     * IN. If referenced Service upon which action is defined is 'partner' Service, action
     * represents a response, i.e. message direction is OUT and MEP is either 'in-out' or
     * 'robust-in-only.
     * <p>
     * Note that in either case, <code>Engine</code> SHOULD NOT propogate unchecked
     * exceptions which MAY occur within application layer. If action represents a request, and
     * a response is expected, i.e. mep is 'robust-in-only' or 'in-out', the exception SHOULD
     * be translated into a <code>SystemFault</code> and returned to caller via the
     * {@link sendAction} method. If action represents a partner response, the error should be
     * logged.
     * 
     * @param actionContext
     *        Defines context within which action occurs. If associated message direction is
     *        IN, action MUST be an action declared on Interface implemented by 'my' Service,
     *        i.e. manifested as an instance of {@link MessageContext}.
     *        <p>
     *        If associated message direction is OUT and action is an output message declared
     *        on Interface implemented by 'partner' Service, action MUST be manifested as an
     *        instance of {@link MessageContext}.
     *        <p>
     *        If associated message direction is OUT and action is a fault message declared on
     *        Interface implemented by 'partner' Service, action MUST be manifested as an
     *        instance of {@link FaultContext} which contains a
     *        {@link BusinessFault#getMessage()}.
     *        <p>
     *        If associated message direction is OUT and action is an undeclared fault message
     *        returned by 'partner' service, action MUST be manifested as an instance of
     *        {@link FaultContext} which contains a {@link SystemFault#getMessage()}.
     * 
     */
    public final void receiveAction(ActionContext actionContext) {
        getRootComponent().receiveAction(this, actionContext);
    }

    /**
     * Convenience method which constructs an instance of undeclared action
     * {@link SystemFault#ACTION}.
     * 
     * @param relatesTo
     *        message id of the message being replied to.
     * @param faultReason
     *        fault description
     * @return FaultContext
     */
    public final FaultContext createSystemFaultAction(String relatesTo, String faultReason) {
        if (relatesTo == null) {
            throw new IllegalArgumentException("relatesTo null");
        }
        if (faultReason == null) {
            throw new IllegalArgumentException("faultReason null");
        }
        ApplicationRT application = getRootComponent().getApplication();
        MessageModules mm = new MessageModules(new WSAMessageModule(SystemFault.ACTION));
        return new FaultContext(mm, new SystemFault(application, faultReason));
    }

    /**
     * Convenience method which constructs an instance of undeclared action
     * {@link SystemFault#ACTION}.
     * 
     * @param relatesTo
     *        message id of the message being replied to.
     * @param faultReason
     *        fault description
     * @param faultCode
     *        A specific indication of fault, i.e. a fault sub-type, within context of the
     *        general classification {@link #NAME}.
     * @return FaultContext
     */
    public final FaultContext createSystemFaultAction(String relatesTo, QName faultCode, String faultReason) {
        if (relatesTo == null) {
            throw new IllegalArgumentException("relatesTo null");
        }
        if (faultCode == null) {
            throw new IllegalArgumentException("faultCode null");
        }
        if (faultReason == null) {
            throw new IllegalArgumentException("faultReason null");
        }
        ApplicationRT application = getRootComponent().getApplication();
        WSAMessageModule wsamm = new WSAMessageModule(SystemFault.ACTION);
        wsamm.setRelatesTo(relatesTo);
        MessageModules mm = new MessageModules(wsamm);
        return new FaultContext(mm, new SystemFault(application, faultCode, faultReason));
    }

    /**
     * Convenience method which constructs an instance of undeclared action
     * {@link SystemFault#ACTION}.
     * 
     * @param relatesTo
     *        message id of the message being replied to.
     * @param faultReason
     *        fault description
     * @param cause
     *        Throwable which caused fault. Note throwable MAY be an instanceof
     *        <code>SystemFault</code>, i.e. a 'nested' fault.
     *        <p>
     *        Note that if <code>Throwable</code> is not an instance of
     *        <code>SystemFault</code>, a stacktrace is serialized to 'FaultDetail' element
     *        of message returned via {@link SystemFault#toMessage()}.
     * @return FaultContext
     */
    public final FaultContext createSystemFaultAction(String relatesTo, String faultReason, Throwable cause) {
        if (relatesTo == null) {
            throw new IllegalArgumentException("relatesTo null");
        }
        if (cause == null) {
            throw new IllegalArgumentException("cause null");
        }
        if (faultReason == null) {
            throw new IllegalArgumentException("faultReason null");
        }
        ApplicationRT application = getRootComponent().getApplication();
        WSAMessageModule wsamm = new WSAMessageModule(SystemFault.ACTION);
        wsamm.setRelatesTo(relatesTo);
        MessageModules mm = new MessageModules(wsamm);
        return new FaultContext(mm, new SystemFault(application, faultReason, cause));
    }

    /**
     * Convenience method which constructs an instance of undeclared action
     * {@link SystemFault#ACTION}.
     * 
     * @param relatesTo
     *        message id of the message being replied to.
     * @param cause
     *        Throwable which caused fault. Note throwable MAY be an instanceof
     *        <code>SystemFault</code>.
     *        <p>
     *        Note that if <code>Throwable</code> is not an instance of
     *        <code>SystemFault</code>, a stacktrace is serialized to 'FaultDetail' element
     *        of message returned via {@link SystemFault#toMessage()}.
     * @return FaultContext
     */
    public final FaultContext createSystemFaultAction(String relatesTo, Throwable cause) {
        if (relatesTo == null) {
            throw new IllegalArgumentException("relatesTo null");
        }
        if (cause == null) {
            throw new IllegalArgumentException("cause null");
        }
        ApplicationRT application = getRootComponent().getApplication();
        WSAMessageModule wsamm = new WSAMessageModule(SystemFault.ACTION);
        wsamm.setRelatesTo(relatesTo);
        MessageModules mm = new MessageModules(wsamm);
        return new FaultContext(mm, new SystemFault(application, cause));
    }

    /**
     * Convenience method which constructs an instance of undeclared action
     * {@link SystemFault#ACTION}.
     * 
     * @param relatesTo
     *        message id of the message being replied to.
     * @param system
     *        fault
     * @return FaultContext
     */
    public final FaultContext createSystemFaultAction(String relatesTo, SystemFault systemFault) {
        if (relatesTo == null) {
            throw new IllegalArgumentException("relatesTo null");
        }
        if (systemFault == null) {
            throw new IllegalArgumentException("systemFault null");
        }
        WSAMessageModule wsamm = new WSAMessageModule(SystemFault.ACTION);
        wsamm.setRelatesTo(relatesTo);
        MessageModules mm = new MessageModules(wsamm);
        return new FaultContext(mm, systemFault);
    }

    /**
     * Sends action to referenced <code>Endpoint</code>. If referenced Service upon which
     * action is defined is 'partner' Service, action represents a request, i.e. direction is
     * IN. If referenced Service upon which action is defined is 'my' Service, action
     * represents a response, i.e. message direction is OUT and MEP is either 'in-out' or
     * 'robust-in-only.
     * <p>
     * Note that this is a 'convenience' method which delegates call to runtime
     * <code>Binding</code> employed by referenced <code>Endpoint</code>. See runtime
     * binding which defines protocol specific behavior.
     * 
     * @param ActionContext
     *        Defines context within which action occurs. If associated message direction is
     *        IN, action MUST be an action declared on Interface implemented by 'partner'
     *        Service, i.e. manifested as an instance of {@link MessageContext}.
     *        <p>
     *        If associated message direction is OUT and action is an output message declared
     *        on Interface implemented by 'my' Service, action MUST be manifested as an
     *        instance of {@link MessageContext}.
     *        <p>
     *        If associated message direction is OUT and action is a fault message declared on
     *        Interface implemented by 'my' Service, action MUST be manifested as an instance
     *        of {@link FaultContext} which contains a {@link BusinessFault#getMessage()}.
     *        <p>
     *        If associated message direction is OUT and action is an undeclared fault message
     *        returned by 'my' service, action MUST be manifested as an instance of
     *        {@link FaultContext} which contains a {@link SystemFault#getMessage()}.
     * 
     * @throws SystemFault
     *         Which MUST be thrown by <code>Binding</code> if action was NOT 'accepted for
     *         processing' by partner. Fault code MUST be {@link SystemFault#DEFAULT_CODE}.
     * 
     * @see BindingRT#sendAction(EndpointReference, ActionContext)
     */
    public final void sendAction(ActionContext actionContext) throws SystemFault {
        Binding binding = endpoint.getBindingReference().getReferencedComponent();
        binding.getRuntimeProvider().sendAction(this, actionContext);
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getParent()
     */
    public final ServiceReference getParent() {
        return parent;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getRootComponent()
     */
    public final EngineRT getRootComponent() {
        return parent.getRootComponent();
    }

    void enableDefaultModules() throws DeploymentException {

        // enable default binding modules

        Binding binding = endpoint.getBindingReference().getReferencedComponent();

        Server server = SystemContext.getContext().getSystem().getServer();
        for (String featureType : binding.getRuntimeProvider().getRequiredBindingFeatures()) {
            Feature feature = server.getFeature(featureType);
            if (feature == null) {
                throw new DeploymentException("BindingFeature '"
                        + featureType
                        + "', required by Binding "
                        + binding.getName()
                        + " employed by Endpoint referenced by component "
                        + getFragmentIdentifier()
                        + ", is undefined.");
            }
            if (!(feature instanceof BindingFeature)) {
                throw new DeploymentException("BindingFeature '"
                        + featureType
                        + "', required by Binding "
                        + binding.getName()
                        + " employed by Endpoint referenced by component "
                        + getFragmentIdentifier()
                        + ", is not an instanceof BindingFeature.");
            }
            BindingModule bm = null;
            try {
                bm = ((BindingFeature)feature).createDefaultBindingModule();
            } catch (FeatureException fe) {
                throw new DeploymentException("Error creating default module which "
                        + " implements BindingFeature '"
                        + featureType
                        + "', required by Binding "
                        + binding.getName()
                        + " employed by Endpoint referenced by component "
                        + getFragmentIdentifier()
                        + ". "
                        + fe);
            }
            if (bm != null) {
                defaultBindingModules.add(bm);
                defaultFeatureModules.add(bm);
            } else {
                throw new DeploymentException("Error creating default module which "
                        + " implements BindingFeature '"
                        + featureType
                        + "', required by Binding "
                        + binding.getName()
                        + " employed by Endpoint referenced by component "
                        + getFragmentIdentifier()
                        + ". BindingFeature returned null.");
            }
        }

        Collections.sort(defaultBindingModules, new RelativeOrderBindingModuleComparator());

        // enable default application modules

        EngineRT engine = getRootComponent();
        for (String featureType : engine.getRequiredApplicationFeatures()) {
            Feature feature = server.getFeature(featureType);
            if (feature == null) {
                throw new DeploymentException("ApplicationFeature '"
                        + featureType
                        + "', required by Engine "
                        + engine.getName()
                        + " employed by Endpoint referenced by component "
                        + getFragmentIdentifier()
                        + ", is undefined.");
            }
            if (!(feature instanceof ApplicationFeature)) {
                throw new DeploymentException("ApplicationFeature '"
                        + featureType
                        + "', required by Engine "
                        + engine.getName()
                        + " employed by Endpoint referenced by component "
                        + getFragmentIdentifier()
                        + ", is not an instanceof ApplicationFeature.");
            }
            ApplicationModule am = null;
            try {
                am = ((ApplicationFeature)feature).createSpringApplicationModule();
            } catch (FeatureException fe) {
                throw new DeploymentException("Error creating default module which "
                        + " implements ApplicationFeature '"
                        + featureType
                        + "', required by Engine "
                        + engine.getName()
                        + " employed by Endpoint referenced by component "
                        + getFragmentIdentifier()
                        + ". "
                        + fe);
            }
            if (am != null) {
                defaultApplicationModules.add(am);
                defaultFeatureModules.add(am);
            } else {
                throw new DeploymentException("Error creating default module which "
                        + " implements ApplicationFeature '"
                        + featureType
                        + "', required by Engine "
                        + engine.getName()
                        + " employed by Endpoint referenced by component "
                        + getFragmentIdentifier()
                        + ". ApplicationFeature returned null.");
            }
        }

        Collections.sort(defaultApplicationModules, new RelativeOrderApplicationModuleComparator());

        // create default transport module

        BindingRT bindingRT = binding.getRuntimeProvider();

        Class<? extends TransportProtocolRT> tt = null;
        Class<?> target = binding.getRuntimeProvider().getClass();
        while (target != null && tt == null) {
            tt = getTransportProtocolRT(target);
            target = target.getSuperclass();
        }

        if (tt == null) {
            throw new DeploymentException("BindingRT class "
                    + bindingRT.getClass().getName()
                    + " MUST implement exactly one TransportProtocolRT.");
        }

        TransportFeature tf = server.getTransportFeature(tt);
        if (tf == null) {
            throw new DeploymentException("TransportFeature which implements TransportProtocolRT "
                    + tt.getName()
                    + " is undefined.");
        }
        try {
            defaultTransportModule = tf.createDefaultTransportModule(this);
            defaultFeatureModules.add(defaultTransportModule);
        } catch (FeatureException fe) {
            throw new DeploymentException("Error installing default module on EPR with fragmentID "
                    + getFragmentIdentifier()
                    + " which implements TransportFeature '"
                    + tf.getExtensionType()
                    + ". "
                    + fe);
        }

        defaultFeatureModules = Collections.unmodifiableSet(defaultFeatureModules);

    }

    /*
     * Used by parent ServiceReference component to add an 'implied' operation reference
     * autogenerated during deployment
     */
    void addImpliedOperationReference(EndpointOperationReference eor) {
        eor.setParent(this);
        operationReferences.put(eor.getEndpointOperationName(), eor);
    }

    void validate() throws DeploymentException {
        for (EndpointOperationReference endpointOperationReference : operationReferences.values()) {
            endpointOperationReference.validate();
        }
    }

    /*
     * Allows binding to validate endpoint policy. This has to be executed after validate
     * method above, because bindings require that feature modules are installed in order to
     * perform their validation
     */
    void validateAgainstBinding() throws DeploymentException {
        Binding binding = endpoint.getBindingReference().getReferencedComponent();
        binding.getRuntimeProvider().validateEndpointReference(this);
    }

    void establishReference() throws DeploymentException {
        if (endpoint != null) {
            return;
        }
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        endpoint = parent.getReferencedComponent().getEndpoint(endpointName);
        if (endpoint == null) {
            throw new DeploymentException("EndpointReference '"
                    + getFragmentIdentifier()
                    + "' is invalid. Endpoint '"
                    + endpointName
                    + "', not defined on Service "
                    + parent.getReferencedComponent().getName()
                    + ".");
        }
        endpointActionReferences = new HashMap<String, EndpointActionReference>();
        if (operationReferences != null) {
            for (EndpointOperationReference operationReference : operationReferences.values()) {
                operationReference.establishReference();
                for (EndpointActionReference ear : operationReference.getActionReferences()) {
                    endpointActionReferences.put(ear.getAction(), ear);
                }
            }
        }
        endpointActionReferences = Collections.unmodifiableMap(endpointActionReferences);
        // force component deployment to deploy runtime binding by looking
        // up runtime providable using a temp component reference. we use
        // runtime binding within validate phase and it must be deployed
        // before it can be used
        getRootComponent().getComponentContext().findReferencedComponent(new BindingReferenceInternal());
    }

    void deriveEffectivePolicy() throws PolicyException {
        PolicyExpression mergedExpression = null;
        if (elementPolicy != null) {
            mergedExpression = elementPolicy.getMergedExpression();
        }
        RoleReference roleReference = parent.getCorrespondingRoleReference();
        if (roleReference.getElementPolicy() != null) {
            PolicyExpression pe = roleReference.getElementPolicy().getMergedExpression();
            mergedExpression = (mergedExpression == null) ? pe : mergedExpression.mergeExpression(pe);
        }
        if (mergedExpression != null) {
            effectivePolicy = new EndpointPolicy(mergedExpression.normalizeExpression());
        }
        if (operationReferences != null) {
            for (EndpointOperationReference operationReference : operationReferences.values()) {
                operationReference.deriveEffectivePolicy();
            }
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.ComponentRT#getFragmentIdentifier()
     */
    public final FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(parent.getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append(endpointName);
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA,
                    "endpointReference", pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(ServiceReference parent) {
        this.parent = parent;
    }

    boolean isAutoGenerated() {
        return isAutoGenerated;
    }

    void setAutoGenerated(boolean isAutoGenerated) {
        this.isAutoGenerated = isAutoGenerated;
    }

    private Class<? extends TransportProtocolRT> getTransportProtocolRT(Class<?> target) {
        for (Class<?> intrface : target.getInterfaces()) {
            if (TransportProtocolRT.class.isAssignableFrom(intrface)) {
                return intrface.asSubclass(TransportProtocolRT.class);
            }
        }
        return null;
    }

    static class RelativeOrderBindingModuleComparator implements Comparator<BindingModule> {
        public int compare(BindingModule o1, BindingModule o2) {
            return o1.getRelativeOrder() - o2.getRelativeOrder();
        }
    }

    private class BindingReferenceInternal implements ComponentReference<Binding> {

        public FragmentIdentifier getFragmentIdentifier() {
            return new FragmentIdentifier("foo", Scheme.EOA, "bar", "");
        }

        public Binding getReferencedComponent() {
            throw new UnsupportedOperationException();
        }

        public ComponentName getReferencedComponentName() {
            return endpoint.getBindingReference().getReferencedComponentName();
        }

        public ComponentType getReferencedComponentType() {
            return ComponentType.BINDING;
        }

        public Class<Binding> getReferencedComponentClass() {
            return Binding.class;
        }

    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("endpointName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(endpointName);
        sb.append(System.getProperty("line.separator"));
        sb.append("operationReferences: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(operationReferences);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(endpointName) + hashCode(operationReferences);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof EndpointReference)) {
            return false;
        }

        if ((endpointName == ((EndpointReference)obj).endpointName || (endpointName != null && endpointName
                .equals(((EndpointReference)obj).endpointName)))
                && (operationReferences == ((EndpointReference)obj).operationReferences || (operationReferences != null && operationReferences
                        .equals(((EndpointReference)obj).operationReferences)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}