/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.engine.rt;

import java.net.URL;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.ApplicationClassLoader;
import org.bluestemsoftware.specification.eoa.DeploymentClassLoader;
import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentContext;
import org.bluestemsoftware.specification.eoa.component.ComponentReference;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.application.Application;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationRT;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationReference;
import org.bluestemsoftware.specification.eoa.component.engine.Engine;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.SystemFault;
import org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT.SingletonRT;
import org.bluestemsoftware.specification.eoa.ext.ExecutableException;
import org.bluestemsoftware.specification.eoa.ext.ExecutableExtension;
import org.bluestemsoftware.specification.eoa.ext.ExtensionException;
import org.bluestemsoftware.specification.eoa.ext.engine.EngineFactory;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyException;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.bluestemsoftware.specification.eoa.system.server.ApplicationFeature;
import org.bluestemsoftware.specification.eoa.system.server.ModulableFeature;

/**
 * Models an <code>Application</code> component at 'runtime' (concrete), i.e. as compared to
 * 'compile-time' (abstract) which is modeled by an instance of {@link ApplicationRT}. More
 * than one runtime application instance MAY exist, each of which is differentiated by engine
 * name.
 * <p>
 * Type specific methods may be accessed by casting <code>EngineRT</code> to a well-known
 * implementation as indicated by {@link #getExtensionType()}. Note that a deployment MUST NOT
 * contain more than one runtime engine definition.
 */
public abstract class EngineRT extends ExecutableExtension implements SingletonRT<Engine> {

    protected FragmentIdentifier fragmentIdentifier;
    protected ComponentName componentName;
    protected Map<QName, EngineReference> engineReferences;
    protected ComponentContext componentContext;
    protected EngineReference myEngineReference;
    protected ApplicationClassLoader applicationClassLoader;

    private Application application;
    private Boolean isDeployed = Boolean.FALSE;

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends ExecutableExtension.Provider {

        public void spi_init() throws ExtensionException;

        public void spi_destroy();

        public void spi_suspend() throws ExecutableException;

        public void spi_resume() throws ExecutableException;

        public void spi_receiveAction(EndpointReference endpointReference, ActionContext actionContext);

        public boolean spi_isSuspended();

    }

    public EngineRT(EngineFactory factory, Provider provider, ComponentName componentName,
            Map<QName, EngineReference> engineReferences) {
        super(factory, provider);
        if (componentName == null) {
            throw new IllegalArgumentException("componentName null");
        } else {
            this.componentName = componentName;
        }
        fragmentIdentifier = new FragmentIdentifier(componentName.getNamespaceURI(), Scheme.EOA, "engineRT",
                componentName.getLocalPart());
        if (engineReferences == null) {
            throw new IllegalArgumentException("engineReferences null");
        } else {
            this.engineReferences = engineReferences;
            for (EngineReference engineReference : engineReferences.values()) {
                engineReference.setParent(this);
            }
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#getName()
     */
    public ComponentName getName() {
        return componentName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#getComponentType()
     */
    public ComponentType getComponentType() {
        return ComponentType.ENGINE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#getComponentContext()
     */
    public ComponentContext getComponentContext() {
        return componentContext;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#getRuntimeProvidable()
     */
    public Engine getRuntimeProvidable() {
        return SystemContext.getContext().getSystem().getEngine(componentName);
    }

    /**
     * Gets classloader required by an <code>Application</code> at runtime, i.e. as modeled
     * by this instance of <code>EngineRT</code>.
     * 
     * @return
     */
    public ApplicationClassLoader getApplicationClassLoader() {
        return applicationClassLoader;
    }

    /**
     * Convenience method used to retrieve runtime <code>Application</code> implemented by
     * this runtime <code>Engine</code> instance.
     * 
     * @return
     */
    public ApplicationRT getApplication() {
        if (application == null) {
            throw new IllegalStateException("Runtime engine definition is not yet deployed.");
        }
        return application.getRuntimeProvider();
    }

    /**
     * Retrieves a specific EngineReference component.
     * 
     * @param engineName
     *        name of referenced <code>Engine</code>
     * 
     * @return EngineReference or <code>null</code> if undefined.
     */
    public EngineReference getEngineReference(QName engineName) {
        if (engineReferences == null) {
            return null;
        }
        return engineReferences.get(engineName);
    }

    /**
     * Retrieves all EngineReference objects. Each EngineReference object corresponds to a Role
     * defined on Application component.
     * 
     * @return Collection of EngineReference objects or empty Collection if none defined
     */
    @SuppressWarnings("unchecked")
    public Collection<EngineReference> getEngineReferences() {
        if (engineReferences == null) {
            return Collections.EMPTY_LIST;
        }
        return engineReferences.values();
    }

    /**
     * Convenience method which retrieves a reference to a service hosted by this engine using
     * name of referenced service.
     * 
     * @param serviceName
     *        name of referenced service
     * @return ReferencedService
     */
    public ServiceReference getServiceReference(QName serviceName) {
        return myEngineReference.getServiceReference(serviceName);
    }

    /**
     * Convenience method which retrieves a reference to a service hosted by this engine using
     * roleName, i.e. which maps to a <code>Role</code> on <code>Application</code>
     * implemented by this <code>Engine</code>.
     * 
     * @param roleName
     * @return
     */
    public ServiceReference getServiceReference(String roleName) {
        return myEngineReference.getServiceReference(roleName);
    }

    /**
     * Convenience method which retrieves all references to services hosted by this engine.
     * 
     * @return Collection of ServiceReference objects
     */
    public Collection<ServiceReference> getServiceReferences() {
        return myEngineReference.getServiceReferences();
    }

    /**
     * Identifies features which MUST be enaged when <code>Engine</code> processes an action.
     * An exception is thrown during validation of <code>Engine</code> if each listed feature
     * is not enabled on every <code>EndpointActionReference</code> nested within runtime
     * <code>Engine</code>.
     * <p>
     * Note that if action is the undeclared action {@link SystemFault#ACTION}, required
     * feature is implemented via 'default' <code>ApplicationModule</code> instance retrieved
     * via {@link ApplicationFeature#createSpringApplicationModule(boolean)}.
     * 
     * @return a <code>Set</code> of type ref URI's each of which identifies a specific
     *         feature type or an empty <code>Set</code> if none required.
     * 
     * @see ApplicationFeature#getExtensionType()
     * @see EndpointReference#getSpringApplicationModules()
     */
    public abstract Set<String> getRequiredApplicationFeatures();

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ManageableExtension#init()
     */
    public void init() throws ExtensionException {
        
        if (applicationClassLoader == null) {
            throw new IllegalStateException("Initialization failed. Application classloader not set.");
        }

        // make app component classes visible, i.e. use application
        // classloader as the context classloader instead of engine
        // factory classloader (which app classloader queries). we

        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(applicationClassLoader);
            ((Provider)provider).spi_init();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ManageableExtension#destroy()
     */
    public void destroy() {

        unInstallFeatureModules();

        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(applicationClassLoader);
            ((Provider)provider).spi_destroy();
        } finally {
            thread.setContextClassLoader(cl);
        }

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ExecutableExtension#suspend()
     */
    public void suspend() throws ExecutableException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(applicationClassLoader);
            ((Provider)provider).spi_suspend();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ExecutableExtension#resume()
     */
    public void resume() throws ExecutableException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(applicationClassLoader);
            ((Provider)provider).spi_resume();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ExecutableExtension#isSuspended()
     */
    public boolean isSuspended() {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(applicationClassLoader);
            return ((Provider)provider).spi_isSuspended();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#deploy(org.bluestemsoftware.specification.eoa.component.ComponentContext)
     */
    public synchronized void deploy(ComponentContext componentContext) throws DeploymentException {
        if (isDeployed == null || isDeployed) {
            return; // circular refs are ok
        }
        isDeployed = null;
        this.componentContext = componentContext;
        establishReferences();
        validate();
        createApplicationClassLoader();
        installFeatureModules();
        isDeployed = Boolean.TRUE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT#isDeployed()
     */
    public synchronized boolean isDeployed() {
        return isDeployed;
    }

    /*
     * invoked by endpoint reference
     */
    void receiveAction(EndpointReference endpointReference, ActionContext actionContext) {

        // TODO: i think we should invoke the application modules here, rather than
        // requiring engine to implement logic.

        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(applicationClassLoader);
            ((Provider)provider).spi_receiveAction(endpointReference, actionContext);
        } finally {
            thread.setContextClassLoader(cl);
        }

    }

    private void establishReferences() throws DeploymentException {

        // retrieve application via an instance of component reference which
        // will force deployment to deploy it if not already deployed, i.e.
        // we rely upon its definition to generate our implied structure

        application = componentContext.findReferencedComponent(new ApplicationReferenceInternal());

        generateImpliedStructure();

        for (EngineReference engineReference : engineReferences.values()) {
            engineReference.establishReference();
        }

    }

    private void installFeatureModules() throws DeploymentException {
        try {
            deriveEffectivePolicy();
        } catch (PolicyException pe) {
            throw new DeploymentException(pe);
        }
        for (EngineReference engineReference : engineReferences.values()) {
            for (ServiceReference serviceReference : engineReference.getServiceReferences()) {
                Iterator<EndpointReference> itr = serviceReference.getEndpointReferences();
                while (itr.hasNext()) {
                    EndpointReference er = itr.next();
                    try {
                        for (EndpointOperationReference eor : er.getOperationReferences()) {
                            for (EndpointActionReference ear : eor.getActionReferences()) {
                                ear.installFeatureModules();
                            }
                        }
                        // enable default modules on endpoint after installing modules
                        // configured via policy, i.e. so that default module does not
                        // override configuration which may exist at endpoint granularity,
                        // e.g. a transport object
                        er.enableDefaultModules();
                    } catch (DeploymentException de) {
                        unInstallFeatureModules();
                        throw de;
                    }
                    // now that feature modules are installed, runtime binding can
                    // validate the endpoint reference component
                    er.validateAgainstBinding();
                }
            }
        }
    }

    /**
     * Creates loader as defined by spec. note that we use our factory loader, and
     * application's component loader.
     */
    protected void createApplicationClassLoader() {
        String classLoaderID = getName().toString();
        URL[] urls = new URL[] {};
        String type = getExtensionType();
        EngineFactory factory = SystemContext.getContext().getSystem().getEngineFactory(type);
        DeploymentClassLoader fl = factory.getFactoryContext().getClassLoader();
        DeploymentClassLoader cl = application.getComponentContext().getClassLoader();
        applicationClassLoader = new ApplicationClassLoader(classLoaderID, urls, fl, cl);
    }

    private void unInstallFeatureModules() {
        for (EngineReference engineReference : engineReferences.values()) {
            for (ServiceReference serviceReference : engineReference.getServiceReferences()) {
                Iterator<EndpointReference> itr = serviceReference.getEndpointReferences();
                while (itr.hasNext()) {
                    EndpointReference er = itr.next();
                    for (EndpointOperationReference eor : er.getOperationReferences()) {
                        for (EndpointActionReference ear : eor.getActionReferences()) {
                            for (FeatureModule fm : ear.getFeatureModules()) {
                                try {
                                    ((ModulableFeature)fm.getFeature()).destroyFeatureModule(fm);
                                } catch (RuntimeException ignore) {
                                }
                            }
                        }
                    }
                    for (FeatureModule fm : er.getDefaultFeatureModules()) {
                        try {
                            ((ModulableFeature)fm.getFeature()).destroyFeatureModule(fm);
                        } catch (RuntimeException ignore) {
                        }
                    }
                }
            }
        }
    }

    /**
     * Used to generate implied structure which MAY be overridden by type specific descendants
     * as necessary.
     * 
     * @param action
     * @return
     */
    protected EngineReference createMyEngineReference(Map<QName, ServiceReference> serviceReferences) {
        return new EngineReference(componentName, serviceReferences, true);
    }

    protected void generateImpliedStructure() throws DeploymentException {

        // if user did not define a reference to self, we create it
        // here

        EngineReference self = getEngineReference(componentName);
        if (self == null) {
            Map<QName, ServiceReference> serviceReferences = new HashMap<QName, ServiceReference>();
            self = createMyEngineReference(serviceReferences);
            self.setParent(this);
            engineReferences.put(componentName, self);
        }
        myEngineReference = self;
        engineReferences = Collections.unmodifiableMap(engineReferences);

    }

    private void validate() throws DeploymentException {

        // attempt to retrieve runtime application provider. user may have defined
        // an engine rt definition, i.e. this object's definition, without defining
        // runtime application provider def. if so, this reports a friendly error

        ApplicationRT applicationRT = null;
        try {
            applicationRT = application.getRuntimeProvider();
        } catch (IllegalStateException ex) {
            throw new DeploymentException("Engine " + getName() + " definition is invalid. " + ex.getMessage());
        }

        for (EngineReference engineReference : engineReferences.values()) {
            try {
                engineReference.validate();
            } catch (DeploymentException de) {
                throw new DeploymentException("Engine reference "
                        + getFragmentIdentifier()
                        + " failed validation. "
                        + de.getMessage());
            }
        }

        // verify that a partner engine is defined for each partner application
        // defined by our application. note that this is not something we can
        // default, i.e. because there may be more than one engine instance
        // which implements the same application

        for (ApplicationReference applicationReference : applicationRT.getApplicationReferences()) {
            QName pa = applicationReference.getReferencedComponentName();
            boolean found = false;
            for (EngineReference engineReference : engineReferences.values()) {
                QName ra = engineReference.getCorrespondingApplicationReference().getReferencedComponentName();
                if (ra.equals(pa)) {
                    found = true;
                    break;
                }
            }
            if (!found) {
                throw new DeploymentException("No EngineReference defined which corresponds"
                        + " to Partner Application "
                        + pa
                        + " defined on Application "
                        + applicationRT.getName()
                        + " implemented by Engine "
                        + getName()
                        + ".");
            }
        }

    }

    /*
     * Aggregates element policy expressions by subject. Merge operation forces resolution of
     * policy refeferences which MAY occur within policy expressions, i.e. references to policy
     * components and or remotely defined expressions
     */
    private void deriveEffectivePolicy() throws PolicyException {
        for (EngineReference engineReference : engineReferences.values()) {
            Iterator<ServiceReference> itr = engineReference.getServiceReferences().iterator();
            while (itr.hasNext()) {
                itr.next().deriveEffectivePolicy();
            }
        }

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.ComponentRT#getFragmentIdentifier()
     */
    public FragmentIdentifier getFragmentIdentifier() {
        return fragmentIdentifier;
    }

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("engineName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(getName());
        sb.append(System.getProperty("line.separator"));
        sb.append("engineReferences: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(engineReferences);
        return sb.toString();
    }

    private class ApplicationReferenceInternal implements ComponentReference<Application> {

        public FragmentIdentifier getFragmentIdentifier() {
            return new FragmentIdentifier("foo", Scheme.EOA, "bar", "");
        }

        public Application getReferencedComponent() {
            throw new UnsupportedOperationException();
        }

        public Class<Application> getReferencedComponentClass() {
            return Application.class;
        }

        public ComponentName getReferencedComponentName() {
            return getRuntimeProvidable().getApplicationReference().getReferencedComponentName();
        }

        public ComponentType getReferencedComponentType() {
            return ComponentType.APPLICATION;
        }

    }

}