/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.engine.rt;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentReference;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationReference;
import org.bluestemsoftware.specification.eoa.component.engine.Actor;
import org.bluestemsoftware.specification.eoa.component.engine.Engine;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT;

/**
 * References an <code>Engine</code> component which is either parent <code>Engine</code>
 * or a 'partner' <code>Engine</code>.
 */
public class EngineReference implements NestedComponentRT<EngineRT, EngineRT>, ComponentReference<Engine> {

    private FragmentIdentifier fragmentIdentifier;
    private EngineRT parent;
    private ComponentName engineName;
    private Engine engine;
    private boolean isMyEngine;
    private Map<QName, ServiceReference> serviceReferences;
    private Map<String, ServiceReference> serviceReferencesByRoleName;

    public EngineReference(ComponentName engineName, Map<QName, ServiceReference> serviceReferences,
            boolean isMyEngine) {
        if (engineName == null) {
            throw new IllegalArgumentException("engineName null");
        } else {
            this.engineName = engineName;
        }
        if (serviceReferences == null) {
            throw new IllegalArgumentException("serviceReferences null");
        } else {
            this.serviceReferences = serviceReferences;
        }
        this.isMyEngine = isMyEngine;
        serviceReferencesByRoleName = new HashMap<String, ServiceReference>();
        for (ServiceReference serviceReference : serviceReferences.values()) {
            serviceReference.setParent(this);
        }
    }

    /**
     * Convenience method with retrieves <code>ApplicationReference</code> defined on
     * <code>Application</code> implemented by parent <code>Engine</code> which corresponds
     * with this <code>EngineReference</code>.
     * 
     * @return
     */
    public final ApplicationReference getCorrespondingApplicationReference() {
        ComponentName cn = engine.getApplicationReference().getReferencedComponentName();
        return parent.getApplication().getApplicationReference(cn);
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponent()
     */
    public final Engine getReferencedComponent() {
        return engine;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentName()
     */
    public final ComponentName getReferencedComponentName() {
        return engineName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentType()
     */
    public final ComponentType getReferencedComponentType() {
        return ComponentType.ENGINE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentClass()
     */
    public final Class<Engine> getReferencedComponentClass() {
        return Engine.class;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getParent()
     */
    public final EngineRT getParent() {
        return parent;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getRootComponent()
     */
    public final EngineRT getRootComponent() {
        return parent;
    }

    /**
     * If true referenced runtime <code>Engine</code> is parent engine.
     * 
     * @return
     */
    public final boolean isMyEngine() {
        return isMyEngine;
    }

    /**
     * Gets a ServiceReference object by name of referenced service.
     * 
     * @param serviceName
     *        name of referenced service
     * @return ReferencedService
     */
    public final ServiceReference getServiceReference(QName serviceName) {
        return serviceReferences.get(serviceName);
    }

    /**
     * Retrieves a <code>ServiceReference</code> component by roleName, i.e. maps to a
     * <code>Role</code> on <code>Application</code> implemented by <code>Engine</code>
     * which identifies interface implemented by referenced <code>Service</code>.
     * 
     * @param roleName
     * @return
     */
    public final ServiceReference getServiceReference(String roleName) {
        return serviceReferencesByRoleName.get(roleName);
    }

    /**
     * Gets all ServiceReference objects.
     * 
     * @return Collection of ServiceReference objects
     */
    public final Collection<ServiceReference> getServiceReferences() {
        return serviceReferences.values();
    }

    void establishReference() throws DeploymentException {
        if (engine != null) {
            return;
        }
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        if (isMyEngine) {
            engine = parent.getRuntimeProvidable();
        } else {
            try {
                engine = getRootComponent().getComponentContext().findReferencedComponent(this);
            } catch (DeploymentException de) {
                throw new DeploymentException("EngineReference "
                        + getFragmentIdentifier()
                        + " defined within deployment "
                        + getRootComponent().getComponentContext().getDeployment().getRef()
                        + " is invalid. "
                        , de);
            }
            if (engine == null) {
                throw new DeploymentException("EngineReference "
                        + getFragmentIdentifier()
                        + " is invalid. Referenced Engine "
                        + engineName
                        + " is undefined.");
            }
        }
        generateImpliedStructure();
        if (serviceReferences != null) {
            for (ServiceReference serviceReference : serviceReferences.values()) {
                serviceReference.establishReference();
            }
        }
        for (ServiceReference serviceReference : serviceReferences.values()) {
            String roleName = serviceReference.getRoleName();
            if (serviceReferencesByRoleName.get(roleName) != null) {
                throw new IllegalArgumentException("roleName defined on ServiceReference "
                        + serviceReference.getReferencedComponentName()
                        + " not unique within context of EngineReference "
                        + engineName
                        + ".");
            } else {
                serviceReferencesByRoleName.put(roleName, serviceReference);
            }
        }
    }

    /**
     * Used to generate implied structure which MAY be overridden by type specific descendants
     * as necessary.
     * 
     * @param serviceName
     * @param isMyService
     * @param elementPolicy
     * @param endpointReferences
     * @return
     * @throws DeploymentException
     */
    protected ServiceReference createServiceReference(ComponentName serviceName, boolean isMyService, ElementPolicy elementPolicy, LinkedHashMap<String, EndpointReference> endpointReferences) throws DeploymentException {
        return new ServiceReference(serviceName, isMyService, elementPolicy, endpointReferences);
    }

    protected void generateImpliedStructure() throws DeploymentException {
        LinkedHashMap<String, EndpointReference> eprs = null;
        ServiceReference sr = null;
        for (Actor actor : engine.getActors()) {
            ComponentName serviceName = actor.getReferencedComponentName();
            if (getServiceReference(serviceName) != null) {
                continue;
            }
            eprs = new LinkedHashMap<String, EndpointReference>();
            sr = createServiceReference(serviceName, isMyEngine, null, eprs);
            sr.setParent(this);
            serviceReferences.put(serviceName, sr);
        }
        serviceReferences = Collections.unmodifiableMap(serviceReferences);
    }

    void validate() throws DeploymentException {
        if (getCorrespondingApplicationReference() == null) {
            throw new DeploymentException("ApplicationReference which references Application "
                    + engine.getApplicationReference().getReferencedComponentName()
                    + " is undefined on Application "
                    + getRootComponent().getApplication().getName()
                    + " implemented by parent Engine.");
        }
        for (ServiceReference serviceReference : serviceReferences.values()) {
            serviceReference.validate();
        }
    }

    void setParent(EngineRT parent) {
        this.parent = parent;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.ComponentRT#getFragmentIdentifier()
     */
    public final FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = parent.getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(rootName.getLocalPart());
            pointerPart.append("/");
            pointerPart.append("xmlns(ns=");
            pointerPart.append(engineName.getNamespaceURI());
            pointerPart.append(")");
            pointerPart.append("ns:" + engineName.getLocalPart());
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA, "engineReference",
                    pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("engineName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(engineName);
        sb.append(System.getProperty("line.separator"));
        sb.append("isMyEngine: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(isMyEngine);
        sb.append(System.getProperty("line.separator"));
        sb.append("serviceReferences: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(serviceReferences);
        return sb.toString();
    }

}
