/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.intrface;

import java.util.Collections;
import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentReference;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceOperation.Style;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage;
import org.bluestemsoftware.specification.eoa.component.message.MessagePart;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.component.types.rt.SchemaDocument;
import org.bluestemsoftware.specification.eoa.component.types.rt.SchemaException;

/**
 * Associates an abstract message definition with a 'normal' message exchanged by parent
 * operation.
 */
public abstract class InterfaceMessageReference extends NestedComponent<Interface, InterfaceOperation> implements ComponentReference<InterfaceMessage>, InterfaceAction, Cloneable {

    protected ComponentName interfaceMessageName;
    protected InterfaceMessage interfaceMessage;
    protected String action;
    protected Direction direction;
    protected ContentModel contentModel;
    protected ElementPolicy elementPolicy;
    protected Map<String, Style> applicableStyles;

    /**
     * Constructs an <code>InterfaceMessageReference</code> instance.
     * 
     * @param interfaceMessageName
     *        Qualified name of <code>InterfaceMessage</code>. If content modeled as ANY or
     *        NONE, <code>null</code> MUST be passed.
     * @param action
     *        A URI which uniquely identifes this action. URI MUST be constructed using rules
     *        defined within WS-Addressing 1.0 specification. URI MUST be unique within the
     *        context of parent Interface component. URI SHOULD be globally unique.
     * @param direction
     *        indicates whether message associated with action is 'received from' a node
     *        (direction is 'In') or 'sent to' a node (direction is 'Out') where 'from' and
     *        'to' are from the perspective of Service(s) which implement parent interface.
     * @param contentModel
     *        defines how message content, i.e. payload, is modeled. If parent
     *        <code>Interface</code> definition was imported via WSDL 1.1, no assertion can
     *        be made regarding payload content, in which case <code>null</code> MUST be
     *        passed.
     * @param elementPolicy
     *        element policy attached to this component or <code>null</code> if none defined.
     */
    public InterfaceMessageReference(ComponentName interfaceMessageName, String action, Direction direction,
            ContentModel contentModel, ElementPolicy elementPolicy) {
        super();
        if (contentModel != null) {
            if (contentModel == ContentModel.ELEMENT) {
                if (interfaceMessageName == null) {
                    throw new IllegalArgumentException("interfaceMessageName null");
                }
            } else {
                if (interfaceMessageName != null) {
                    throw new IllegalArgumentException("interfaceMessageName must be null");
                }
            }
        }
        if (action == null) {
            throw new IllegalArgumentException("action null");
        }
        if (direction == null) {
            throw new IllegalArgumentException("direction null");
        }
        this.interfaceMessageName = interfaceMessageName;
        this.action = action;
        this.direction = direction;
        this.contentModel = contentModel;
        this.elementPolicy = elementPolicy;
    }

    /**
     * Defines how message content, i.e. payload, is modeled. The possible values are as
     * follows:
     * <p>
     * ANY Indicates message content is any single element. This value is NOT backwards
     * compatible with WSDL 1.1. If specified, parent Interface cannot be descrided via WSDL
     * 1.1.
     * <p>
     * NONE Indicates message content is empty. This value is NOT backwards compatible with
     * WSDL 1.1. If specified, parent Interface cannot be descrided via WSDL 1.1 which requires
     * a named wsdl:message component.
     * <p>
     * ELEMENT Indicates message content consists of a single element. This value is backwards
     * compatible with WSDL 1.1, i.e. parent Interface can be described via WSDL 1.1. No
     * assertion can be made by client, however, regarding how element is serialized within
     * wire formatted message unless WSDL 1.1 Interface is bound by a WSDL 1.1 Binding.
     */
    public static enum ContentModel {

        ANY, NONE, ELEMENT, OTHER;

        public static final String ANY_TOKEN = "#any";
        public static final String NONE_TOKEN = "#none";
        public static final String ELEMENT_TOKEN = "#element";
        public static final String OTHER_TOKEN = "#other";

        public static ContentModel valueOfToken(String token) {
            if (token == null) {
                return null;
            }
            if (token.equals(ANY_TOKEN)) {
                return ANY;
            }
            if (token.equals(NONE_TOKEN)) {
                return NONE;
            }
            if (token.equals(ELEMENT_TOKEN)) {
                return ELEMENT;
            }
            if (token.equals(OTHER_TOKEN)) {
                return OTHER;
            }
            return null;
        }

        public static String toToken(ContentModel contentModel) {
            if (contentModel == null) {
                return null;
            }
            if (contentModel == ANY) {
                return ANY_TOKEN;
            }
            if (contentModel == NONE) {
                return NONE_TOKEN;
            }
            if (contentModel == ELEMENT) {
                return ELEMENT_TOKEN;
            }
            if (contentModel == OTHER) {
                return OTHER_TOKEN;
            }
            return null;
        }

    }

    /**
     * Returns enumerated value indicating how message content, i.e. payload, is modeled.
     * 
     * @return Enumerated value or <code>null</code> if no assertion can be made regarding
     *         payload content, i.e. parent <code>Interface</code> definition was imported
     *         via WSDL 1.1 which relies upon a WSDL 1.1 <code>Binding</code> to infer the
     *         content model. If <code>null</code>, parent Interface cannot be described via
     *         WSDL 2.0.
     */
    public final ContentModel getContentModel() {
        return contentModel;
    }

    /**
     * Identifies name of referenced <code>Component</code> used to fulfill
     * <code>ComponentReference</code> contract.
     * 
     * @return name of referenced <code>Component</code> or <code>null</code> if content
     *         modeled as ANY, NONE or OTHER.
     */
    public final ComponentName getReferencedComponentName() {
        return interfaceMessageName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentType()
     */
    public ComponentType getReferencedComponentType() {
        return ComponentType.MESSAGE;
    }

    /**
     * Returns referenced <code>Component</code>.
     * 
     * @return InterfaceMessage object or <code>null</code> if content modeled as ANY, NONE
     *         or OTHER.
     */
    public InterfaceMessage getReferencedComponent() {
        return interfaceMessage;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentClass()
     */
    public Class<InterfaceMessage> getReferencedComponentClass() {
        return InterfaceMessage.class;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction#getAction()
     */
    public String getAction() {
        return action;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction#getDirection()
     */
    public Direction getDirection() {
        return direction;
    }

    final void validate() throws DeploymentException {
    }

    /**
     * Gets <code>ElementPolicy</code> attached to component.
     * 
     * @return ElementPolicy or <code>null</code> if no policy attached.
     */
    public final ElementPolicy getElementPolicy() {
        return elementPolicy;
    }

    /**
     * Gets collection of modeled styles applicable to this action. Note that each modeled
     * style MUST be a style defined on parent operation and/or on parent interface.
     * 
     * @return A collection of modeled styles, each of which is mapped to IRI which identifies
     *         style, or an empty collection if none defined.
     * 
     * @see InterfaceOperation#getStyles()
     * @see Interface#getStyleDefaults()
     * @see SchemaDocument#modelStyles(InterfaceAction)
     */
    @SuppressWarnings("unchecked")
    public Map<String, Style> getApplicableStyles() {

        if (applicableStyles == null) {
            return Collections.EMPTY_MAP;
        }

        return applicableStyles;
    }

    /*
     * invoked after provider definitions have been deployed via parent interface, i.e. we need
     * runtime schema to generate the models
     */
    final void modelApplicableStyles() throws DeploymentException {

        // create styles. note that styles are a wsdl 2.0 artifact which implies that
        // message must be a single part wsdl 2.0 style message. if content model has
        // a value, then it's a wsdl 2.0 style message. content must be modeled as
        // an element

        if (interfaceMessage != null & contentModel == ContentModel.ELEMENT) {
            MessagePart part = interfaceMessage.getPart(MessagePart.PAYLOAD);
            try {
                applicableStyles = part.getReferencedComponent().getRuntimeProvider().modelStyles(this);
            } catch (SchemaException se) {
                throw new DeploymentException("Action "
                        + getFragmentIdentifier()
                        + " is invalid. "
                        + se.getMessage());
            }
        }

    }

    final void establishReference() throws DeploymentException {
        if (interfaceMessage != null) {
            return;
        }
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        if (interfaceMessageName == null) {
            return;
        }
        if (interfaceMessageName != null) {
            interfaceMessage = getRootComponent().getComponentContext().findReferencedComponent(this);
            if (interfaceMessage == null) {
                throw new DeploymentException("ComponentReference '"
                        + getFragmentIdentifier()
                        + "' is invalid. Referenced message "
                        + interfaceMessageName
                        + " is undefined.");
            }
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     */
    public final FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(((InterfaceOperation)parent).getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append(direction == Direction.IN ? "In" : "Out");
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA,
                    "interfaceMessageReference", pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    @Override
    public Object clone() throws CloneNotSupportedException {
        InterfaceMessageReference imr = (InterfaceMessageReference)super.clone();
        if (elementPolicy != null) {
            imr.elementPolicy = (ElementPolicy)elementPolicy.clone();
        }
        imr.parent = null;
        return imr;
    }

    void setParent(InterfaceOperation parent) {
        this.parent = parent;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("interfaceMessageName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(interfaceMessageName);
        sb.append(System.getProperty("line.separator"));
        sb.append("action: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(action);
        sb.append(System.getProperty("line.separator"));
        sb.append("direction: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(direction);
        sb.append(System.getProperty("line.separator"));
        sb.append("contentModel: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(contentModel);
        return sb.toString();
    }

    @Override
    public final int hashCode() {
        return hashCode(interfaceMessageName) + hashCode(action) + hashCode(direction) + hashCode(contentModel);
    }

    @Override
    public final boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof InterfaceMessageReference)) {
            return false;
        }

        if ((interfaceMessageName == ((InterfaceMessageReference)obj).interfaceMessageName || (interfaceMessageName != null && interfaceMessageName
                .equals(((InterfaceMessageReference)obj).interfaceMessageName)))
                && (action == ((InterfaceMessageReference)obj).action || (action != null && action
                        .equals(((InterfaceMessageReference)obj).action)))
                && (direction == ((InterfaceMessageReference)obj).direction || (direction != null && direction
                        .equals(((InterfaceMessageReference)obj).direction)))
                && (contentModel == ((InterfaceMessageReference)obj).contentModel || (contentModel != null && contentModel
                        .equals(((InterfaceMessageReference)obj).contentModel)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}