/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.intrface.rt;

import java.util.ArrayList;
import java.util.List;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.ComponentDeployment;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.RootComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationRT;
import org.bluestemsoftware.specification.eoa.component.binding.rt.AcceptedResponseError;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage;
import org.bluestemsoftware.specification.eoa.component.message.rt.Content;
import org.bluestemsoftware.specification.eoa.component.message.rt.Message;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

/**
 * Signifies the occurrence of a 'system' related error which occured while processing an
 * <code>Action</code>. Note that unlike business faults, system faults are typically
 * undeclared.
 * 
 * @see FaultContext
 */
public final class SystemFault extends RuntimeException implements ApplicationFault {

    private static final long serialVersionUID = 1L;

    /**
     * Name which identifies wsdl fault name of system fault.
     */
    public static final QName FAULT_NAME = new QName("http://bluestemsoftware.org/specification/eoa",
            "SystemFault");

    /**
     * Identifies semantics of message associated with this fault.
     */
    public static final String ACTION = "http://bluestemsoftware.org/specification/eoa/system/fault";

    /**
     * A generalized code which MAY be used in the absence of a more specific indication of
     * fault.
     */
    public static final QName DEFAULT_CODE = new QName("http://bluestemsoftware.org/specification/eoa",
            "SystemError");

    /**
     * Value used when no specific reason provided.
     */
    public static final String DEFAULT_REASON = "A system related error occurred while processing message.";
    
    /**
     * Name of system fault abstract message definition.
     */
    public static final QName MESSAGE_NAME = new QName(RootComponent.EOA_COMPONENT_NS, "SystemFault");
    
    /**
     * Fragment identifier of system fault abstract message definition.
     */
    public static final FragmentIdentifier MESSAGE_FRAGMENT_ID = new FragmentIdentifier(MESSAGE_NAME
            .getNamespaceURI(), Scheme.EOA, "message", MESSAGE_NAME.getLocalPart());
    
    /**
     * Name of property which MAY be used to extract fault code value from system fault message
     * instance.
     * 
     * @see Message#evaluateString(QName)
     */
    public static final QName FAULT_CODE_PROPERTY = new QName(RootComponent.EOA_COMPONENT_NS, "SystemFaultCode");

    /**
     * Name of property which MAY be used to extract fault reason text from system fault message
     * instance.
     * 
     * @see Message#evaluateString(QName)
     */
    public static final QName FAULT_REASON_PROPERTY = new QName(RootComponent.EOA_COMPONENT_NS, "SystemFaultReason");

    /**
     * Name of property which MAY be used to extract fault detail element from system fault message
     * instance.
     * 
     * @see Message#evaluateNode(QName)
     */
    public static final QName FAULT_DETAIL_PROPERTY = new QName(RootComponent.EOA_COMPONENT_NS, "SystemFaultDetail");
    
    /**
     * Name of property which MAY be used to extract fault diagnostic element from system fault message
     * instance.
     * 
     * @see Message#evaluateNode(QName)
     */
    public static final QName FAULT_DIAGNOSTIC_PROPERTY = new QName(RootComponent.EOA_COMPONENT_NS, "SystemFaultDiagnostic");
    
    /**
     * Name of property which MAY be used to extract fault cause element from system fault message
     * instance.
     * 
     * @see Message#evaluateNode(QName)
     */
    public static final QName FAULT_CAUSE_PROPERTY = new QName(RootComponent.EOA_COMPONENT_NS, "SystemFaultCause");
       
    private ApplicationRT application;
    private QName faultCode = DEFAULT_CODE;
    private List<Content> faultDetail;
    private Message faultMessage;
    
    private static InterfaceMessage abstractMessage;

    /**
     * Constructs an instance of <code>SystemFault</code> with given reason and implied fault
     * code {@link #DEFAULT_CODE}.
     * 
     * @param application
     *        runtime <code>Application</code> instance within which fault occurred.
     * @param faultReason
     *        fault description
     */
    public SystemFault(ApplicationRT application, String faultReason) {
        super(faultReason);
        if (application == null) {
            throw new IllegalArgumentException("application null");
        } else {
            this.application = application;
        }
    }

    /**
     * Constructs an instance of <code>SystemFault</code> with given code and given reason.
     * 
     * @param application
     *        runtime <code>Application</code> instance within which fault occurred.
     * @param faultCode
     *        A specific indication of fault, i.e. a fault sub-type, within context of the
     *        general classification {@link #NAME}.
     * @param faultReason
     *        fault description
     */
    public SystemFault(ApplicationRT application, QName faultCode, String faultReason) {
        super(faultReason);
        if (application == null) {
            throw new IllegalArgumentException("application null");
        } else {
            this.application = application;
        }
        if (faultCode == null) {
            throw new IllegalArgumentException("faultCode null");
        } else {
            this.faultCode = faultCode;
        }
    }

    /**
     * Constructs an instance of <code>SystemFault</code>.
     * 
     * @param application
     *        runtime <code>Application</code> instance within which fault occurred.
     * @param faultReason
     *        fault description
     * @param faultCode
     *        A specific indication of fault, i.e. a fault sub-type, within context of the
     *        general classification {@link #NAME}.
     * @param faultDetail
     *        or <code>null</code> if none defined
     */
    public SystemFault(ApplicationRT application, QName faultCode, String faultReason, List<Content> faultDetail) {
        super(faultReason);
        if (application == null) {
            throw new IllegalArgumentException("application null");
        } else {
            this.application = application;
        }
        if (faultCode == null) {
            throw new IllegalArgumentException("faultCode null");
        } else {
            this.faultCode = faultCode;
        }
        this.faultDetail = faultDetail;
    }

    /**
     * Constructs an instance of <code>SystemFault</code> with given reason, given cause, and
     * implied fault code {@link #DEFAULT_CODE}.
     * 
     * @param application
     *        runtime <code>Application</code> instance within which fault occurred.
     * @param faultReason
     *        fault description
     * @param cause
     *        Throwable which caused fault. Note throwable MAY be an instanceof
     *        <code>SystemFault</code>, i.e. a 'nested' fault.
     *        <p>
     *        Note that if <code>Throwable</code> is not an instance of
     *        <code>SystemFault</code>, a stacktrace is serialized to 'FaultDetail' element
     *        of message returned via {@link #toMessage()}.
     * 
     */
    public SystemFault(ApplicationRT application, String faultReason, Throwable cause) {
        super(faultReason, cause);
        if (application == null) {
            throw new IllegalArgumentException("application null");
        } else {
            this.application = application;
        }
    }

    /**
     * Constructs an instance of <code>SystemFault</code> with given cause, default reason
     * {@link #DEFAULT_REASON} and implied fault code {@link #DEFAULT_CODE}.
     * 
     * @param application
     *        runtime <code>Application</code> instance within which fault occurred.
     * @param cause
     *        Throwable which caused fault. Note throwable MAY be an instanceof
     *        <code>SystemFault</code>, i.e. a 'nested' fault.
     *        <p>
     *        Note that if <code>Throwable</code> is not an instance of
     *        <code>SystemFault</code>, a stacktrace is serialized to 'FaultDetail' element
     *        of message returned via {@link #toMessage()}.
     * 
     */
    public SystemFault(ApplicationRT application, Throwable cause) {
        super(DEFAULT_REASON, cause);
        if (application == null) {
            throw new IllegalArgumentException("application null");
        } else {
            this.application = application;
        }
    }

    /**
     * Returns the most specific indication of fault, i.e. a fault sub-type, within context of
     * the general classification {@link #NAME}.
     * <p>
     * Note that while applications are encouraged to develop a set of application specific
     * sub-types, the default value {@link #DEFAULT_CODE} MAY be used.
     * 
     * @return fault code
     */
    public final QName getFaultCode() {
        return faultCode;
    }

    /**
     * Returns fault reason. Note, returns same value as {@link #getMessage()}.
     * 
     * @return fault reason
     */
    public final String getFaultReason() {
        return getMessage();
    }

    /**
     * Returns a <code>List</code> of elements which MAY provide additional context regarding
     * the specifics of fault.
     * 
     * @return List of detail elements, or <code>null</code> if none defined.
     */
    public final List<Content> getFaultDetail() {
        return faultDetail;
    }

    /**
     * Returns name of implied InterfaceFault instance which, in this case, is the static value
     * {@link #FAULT_NAME}.
     * 
     * @return fault name
     */
    public final QName getFaultName() {
        return FAULT_NAME;
    }
    
    /**
     * Gets system fault's built-in abstract message definition.
     * @return
     */
    public static InterfaceMessage getAbstractMessage() {
        if (abstractMessage == null) {
            abstractMessage = SystemContext.getContext().getSystem().getMessage(MESSAGE_NAME);
            if (abstractMessage == null) {
                throw new IllegalStateException("Built-in message " + MESSAGE_NAME + " is undefined");
            }
        }
        return abstractMessage;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.intrface.rt.ApplicationFault#toMessage()
     */
    public Message toMessage() {
        if (faultMessage != null) {
            return faultMessage;
        } else {
            faultMessage = toMessage(this);
        }
        return faultMessage;
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Throwable#toString()
     */
    @Override
    public String toString() {
        return toMessage().toString();
    }

    private Message toMessage(SystemFault systemFault) {

        InterfaceMessage am = getAbstractMessage();
        ComponentDeployment deployment = application.getComponentContext().getDeployment();
        Message answer = (Message)deployment.getProvider(ComponentType.MESSAGE, am.getName());
        if (answer == null) {
            answer = ((Message)am.getRuntimeProvider()).cloneDefinition();
        }

        Content root = answer.createContent(MESSAGE_NAME.getNamespaceURI(), "eoa:SystemFault");
        answer.setContent(root);

        Element faultCode = answer.createElementNS(MESSAGE_NAME.getNamespaceURI(), "eoa:FaultCode");
        root.appendChild(faultCode);
        String prefix = null;
        try {
            prefix = root.lookupPrefix(systemFault.getFaultCode().getNamespaceURI());
        } catch (UnsupportedOperationException ex) {
            // not yet implemented on axiom dom (as of version 1.2.5)
        }
        if (prefix == null) {
            prefix = "ns";
            faultCode.setAttributeNS("http://www.w3.org/2000/xmlns/", "xmlns:ns", systemFault.getFaultCode()
                    .getNamespaceURI());
        }
        faultCode.appendChild(answer.createTextNode(prefix + ":" + systemFault.getFaultCode().getLocalPart()));

        Element faultReason = answer.createElementNS(MESSAGE_NAME.getNamespaceURI(), "eoa:FaultReason");
        root.appendChild(faultReason);
        faultReason.appendChild(answer.createTextNode(systemFault.getFaultReason()));

        // if fault detail is specified, cause was not manifested as a throwable,
        // i.e. it's described via detail. append the detail and we're done

        if (systemFault.getFaultDetail() != null) {
            Element temp = answer.createElementNS(MESSAGE_NAME.getNamespaceURI(), "eoa:FaultDetail");
            root.appendChild(temp);
            for (Content detail : systemFault.getFaultDetail()) {
                temp.appendChild(answer.importNode(detail, true));
            }
            return answer;
        }

        // if fault detail is null and cause is null, we're done. return the
        // constructed message

        if (systemFault.getCause() == null) {
            return answer;
        }

        // if cause is an instance of system fault, we embed fault message instance
        // recursively. otherwise we write stack trace to detail

        Element temp = answer.createElementNS(MESSAGE_NAME.getNamespaceURI(), "eoa:FaultDetail");
        root.appendChild(temp);

        if (systemFault.getCause() instanceof SystemFault) {
            Message cause = toMessage((SystemFault)systemFault.getCause());
            temp.appendChild(answer.importNode(cause.getContent(), true));
        } else {
            Element ste = answer.createElementNS(RootComponent.EOA_COMPONENT_NS, "eoa:FaultDiagnostic");
            temp.appendChild(ste);
            StringBuilder sb = new StringBuilder("<![CDATA[");
            sb.append(getCause().toString());
            sb.append("]]>");
            ste.appendChild(answer.createTextNode(sb.toString()));
        }

        return answer;

    }

    /**
     * Converts an <code>AcceptedResponseError</code> into a <code>SystemFault</code>.
     * Used by <code>Binding</code> to communicate a fault to <code>Application</code>
     * which occurred within 'binding layer' after partner response was 'accepted for
     * processing' by binding.
     * 
     * @param application
     *        to which fault will be communicated
     * @param bindingFault
     * @return SystemFault
     */
    public static SystemFault valueOf(ApplicationRT application, AcceptedResponseError acceptedResponseError) {
        if (application == null) {
            throw new IllegalArgumentException("application null");
        }
        if (acceptedResponseError == null) {
            throw new IllegalArgumentException("bindingFault null");
        }
        InterfaceMessage am = getAbstractMessage();
        ComponentDeployment deployment = application.getComponentContext().getDeployment();
        Message message = (Message)deployment.getProvider(ComponentType.MESSAGE, am.getName());
        if (message == null) {
            message = ((Message)am.getRuntimeProvider()).cloneDefinition();
        }
        List<Content> detail = new ArrayList<Content>();
        if (acceptedResponseError.getDetail() != null) {
            List<Content> temp = acceptedResponseError.getDetail();
            for (Content content : temp) {
                Node imported = null;
                try {
                    imported = message.importContent(content);
                } catch (UnsupportedOperationException ue) {
                    imported = message.importNode(content, true);
                }
                detail.add((Content)imported);
            }
        }
        return new SystemFault(application, acceptedResponseError.getFaultName(), acceptedResponseError
                .getFaultReason(), detail);
    }

    /**
     * Constructs <code>SystemFault</code> from a runtime instance of built-in system fault
     * message. {@link #MESSAGE_NAME}.
     * 
     * @param application
     *        within which the generated fault occurred
     * @param message
     */

    public static SystemFault valueOf(ApplicationRT application, Message message) {
        if (message == null) {
            throw new IllegalArgumentException("message null");
        }
        if (!message.getName().equals(MESSAGE_NAME)) {
            throw new IllegalArgumentException("Message MUST be a runtime instance of " + MESSAGE_NAME);
        }
        String faultReason = null;
        QName faultCode = null;
        NodeList nodeList = message.getContent().getChildNodes();
        for (int i = 0; i < nodeList.getLength() && faultReason == null && faultCode == null; i++) {
            if (nodeList.item(i).getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }
            Element element = (Element)nodeList.item(i);
            if (element.getNamespaceURI().equals(MESSAGE_NAME.getNamespaceURI())) {
                if (element.getLocalName().equals("FaultCode")) {
                    String prefixedName = ((Text)element.getFirstChild()).getData();
                    faultCode = ComponentName.valueOf(element, prefixedName);
                }
                if (element.getLocalName().equals("FaultReason")) {
                    faultReason = ((Text)element.getFirstChild()).getData();
                }
            }
        }
        SystemFault sf = new SystemFault(application, faultCode, faultReason);
        sf.faultMessage = message;
        return sf;
    }

}
