/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.message;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.RootComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.RuntimeProvidable;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationRT;
import org.bluestemsoftware.specification.eoa.component.message.rt.Message;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT;
import org.bluestemsoftware.specification.eoa.component.types.SchemaComponent;
import org.bluestemsoftware.specification.eoa.component.types.SchemaFragment;
import org.bluestemsoftware.specification.eoa.component.types.SchemaProperty;
import org.bluestemsoftware.specification.eoa.component.types.SchemaComponent.Type;

/**
 * Models an abstract message component. Models a WSDL 2.0 document style message, while
 * providing backwards compatibility with WSDL 1.1 multi-part messages.
 * <p>
 * Note that use of abstract messages with more than one part is discouraged. Applications
 * should be designed to receive/send abstract messages with a single part whose content, i.e.
 * payload, is modeled as an element. Abstract message with more than one part cannot be
 * described via WSDL 2.0.
 */

public class InterfaceMessage extends RootComponent implements RuntimeProvidable {

    boolean referencesEstablished;
    private ElementPolicy elementPolicy;
    private Map<String, MessagePart> messageParts;
    private Map<QName, MessageFragment> messageFragments;

    /**
     * Constructs an abstract message component.
     * 
     * @param name
     *        Qualified name of abstract message. Note that if interface definition is read
     *        from a WSDL 2.x description, value of 'element' attribute defined on wsdl
     *        input/output/fault MUST be used as a proxy for abstract message name.
     * @param elementPolicy
     *        element policy attached to this component.
     * @param messageParts
     *        Map of abstract message part components, each of which is mapped to message part
     *        name.
     */
    public InterfaceMessage(ComponentName name, ElementPolicy elementPolicy, Map<String, MessagePart> messageParts) {

        super();

        if (name == null) {
            throw new IllegalArgumentException("name null");
        } else {
            this.name = name;
        }
        fragmentIdentifier = new FragmentIdentifier(name.getNamespaceURI(), Scheme.EOA, "message", name
                .getLocalPart());
        if (messageParts == null) {
            throw new IllegalArgumentException("messageParts null");
        } else {
            this.messageParts = Collections.unmodifiableMap(messageParts);
        }
        this.elementPolicy = elementPolicy;
        for (MessagePart messagePart : messageParts.values()) {
            messagePart.setParent(this);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#getComponentType()
     */
    @Override
    public final ComponentType getComponentType() {
        return ComponentType.MESSAGE;
    }

    /**
     * Returns 'default' runtime message instance which MUST be used by a compliant
     * <code>Binding</code> and/or <code>Engine</code> when an <code>Application</code>
     * fails to define a 'scoped' provider definition.
     * 
     * @return Message default provider instance which MUST be cloned before it can be used
     * 
     * @see Message#clone(boolean)
     * @see ApplicationRT#getMessageProvider(QName)
     */
    public RootComponentRT getRuntimeProvider() {
        if (runtimeProvider == null || !runtimeProvider.isDeployed()) {
            throw new IllegalStateException("Provider definition not yet deployed.");
        }
        return (Message)runtimeProvider;
    }

    /**
     * Convenience method which indicates whether metadata describes a WSDL 2.0 style abstract
     * message.
     * 
     * @return true if message is empty, i.e. no abstract parts defined, or if message has
     *         exactly one part typed as an element.
     */
    public boolean isWSDL20Style() {
        if (messageParts.size() == 0) {
            return true;
        }
        if (messageParts.size() == 1) {
            Type partType = messageParts.values().iterator().next().getSchemaComponentType();
            if (partType == Type.ELEMENT_DECLARATION) {
                return true;
            }
        }
        return false;
    }

    /**
     * Gets abstract part component.
     * 
     * @param partName
     *        Name of part.
     * @return MessagePart Metadata object describing indicated message part, or null if part
     *         not found.
     */
    public final MessagePart getPart(String partName) {
        return messageParts.get(partName);
    }

    /**
     * Gets metadata describing all message parts defined on this abstract message.
     * <p>
     * 
     * @return Collection of MessagePart objects, or an empty collection if no parts defined.
     */
    public final Collection<MessagePart> getParts() {
        return messageParts.values();
    }

    /**
     * Retrieves a message fragment definition. A message fragment is used to extract a runtime
     * instance of {@link SchemaProperty} embedded within a message part instance.
     * 
     * @param QName
     *        Qualified name of fragment Type. Note that fragment type is unique within context
     *        of message which implies that only one instance of embedded Type is allowed per
     *        message.
     * @return MessageFragment fragment component.
     */
    public final MessageFragment getFragment(QName fragmentName) {
        if (messageFragments != null) {
            return messageFragments.get(fragmentName);
        } else {
            return null;
        }
    }

    /**
     * Retrieves all message fragment definitions. A message fragment is used to extract a
     * runtime instance of {@link SchemaProperty} embedded within a message part instance.
     * <p>
     * For each abstract message part defined on parent message, a compliant implementation
     * MUST define an <code>MessageFragment</code> for each schema fragment nested within
     * referened <code>SchemaComponent</code>.
     * 
     * @return Collection of message fragment components or an empty collection if none
     *         defined.
     */
    @SuppressWarnings("unchecked")
    public final Collection<MessageFragment> getFragments() {
        if (messageFragments != null) {
            return messageFragments.values();
        } else {
            return Collections.EMPTY_SET;
        }
    }

    /**
     * Gets <code>ElementPolicy</code> attached to component.
     * <p>
     * Note that this attachment point is provided for backwards compatibility with WSDL 1.1,
     * i.e. for message definitions imported via a WSDL 1.1 defintion. If specified, policy
     * cannot be described via WSDL 2.x. Its use is therefore discouraged.
     * 
     * @return elementPolicy or <code>null</code> if no attached policy
     */
    public final ElementPolicy getElementPolicy() {
        return elementPolicy;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#validate()
     */
    @Override
    protected final void validate() throws DeploymentException {
        
        // TODO: implement
        
        // note. no need to validate runtime provider definition. it validates
        // itself when deployed, i.e. because more than one instance may exist
        
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#establishReferences()
     */
    @Override
    protected synchronized void establishReferences() throws DeploymentException {
        if (referencesEstablished) {
            return;
        }
        messageFragments = new HashMap<QName, MessageFragment>();
        for (MessagePart part : messageParts.values()) {
            part.establishReference();
            SchemaComponent schemaComponent = part.getSchemaComponent();
            Iterator<?> itr = schemaComponent.getFragments().iterator();
            while(itr.hasNext()) {
                SchemaFragment fragment = (SchemaFragment)itr.next();
                QName propertyName = fragment.getSchemaPropertyQName();
                if (messageFragments.containsKey(propertyName)) {
                    throw new DeploymentException("Error establishing reference on abstract message part'"
                            + part.getName()
                            + "'. A message fragment used to extract embedded SchemaProperty instance "
                            + propertyName
                            + " is already defined on abstract message '"
                            + name
                            + "'. Message fragment name, i.e. qualified name of embedded property,"
                            + " MUST be unique within context of message.");
                }
                MessageFragment mf = new MessageFragment(propertyName, part.getName(), fragment);
                mf.setParent(this);
                messageFragments.put(propertyName, mf);
            }
        }
        referencesEstablished = true;
    }

    public static final class AnyMessage extends InterfaceMessage {

        public static final ComponentName NAME = new ComponentName(RootComponent.EOA_COMPONENT_NS, "AnyMessage");

        private static AnyMessage singleton = null;

        private AnyMessage(ComponentName name, ElementPolicy elementPolicy, Map<String, MessagePart> messageParts) {
            super(name, elementPolicy, messageParts);
        }

        public static final AnyMessage getInstance() {
            if (singleton == null) {
                Map<String, MessagePart> messageParts = new HashMap<String, MessagePart>();
                messageParts.put(MessagePart.PAYLOAD, new MessagePart(MessagePart.PAYLOAD, "http://www.w3.org/2001/XMLSchema",
                        Type.ELEMENT_DECLARATION, "any"));
                singleton = new AnyMessage(NAME, null, messageParts);
            }
            return singleton;
        }

    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(getName());
        sb.append(System.getProperty("line.separator"));
        sb.append("parts: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(messageParts);
        sb.append(System.getProperty("line.separator"));
        sb.append("messageFragments: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(messageFragments);
        return sb.toString();
    }

    @Override
    public final int hashCode() {
        return hashCode(getName()) + hashCode(messageParts) + hashCode(messageFragments);
    }

    @Override
    public final boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof InterfaceMessage)) {
            return false;
        }

        if ((getName() == ((InterfaceMessage)obj).getName() || (getName() != null && getName().equals(
                ((InterfaceMessage)obj).getName())))

        ) {
            if (messageParts == ((InterfaceMessage)obj).messageParts
                    || (messageParts != null && messageParts.equals(((InterfaceMessage)obj).messageParts))) {
                if (messageFragments == ((InterfaceMessage)obj).messageFragments
                        || (messageFragments != null && messageFragments
                                .equals(((InterfaceMessage)obj).messageFragments))) {
                    return true;
                }
            }
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}