/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.message;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.types.SchemaFragment;

/**
 * Messages may contain named fragments, i.e. embedded 'pieces' of runtime data, each of which
 * is modeled by a referenced schema fragment, i.e. a {@link SchemaFragment}.
 * <p>
 * A <code>MessageFragment</code> is automatically generated for each
 * <code>SchemaFragment</code> defined on each <code>SchemaComponent</code> referenced by
 * each abstract message part defined on parent <code>InterfaceMessage</code>.
 */
public final class MessageFragment extends NestedComponent<InterfaceMessage, InterfaceMessage> {

    private QName name;
    private String partName;
    private SchemaFragment schemaFragment;

    /**
     * Constructs an <code>MessageFragment</code> component.
     * 
     * @param name
     *        message fragment name. A compliant implementation MUST use property name defined
     *        on referenced schema fragment as a proxy for fragment name. Name MUST be unique
     *        within context of parent message.
     * @param partName
     *        name of abstract message part which contains fragment.
     */
    public MessageFragment(QName name, String partName, SchemaFragment schemaFragment) {
        if (name == null) {
            throw new IllegalArgumentException("name null");
        } else {
            this.name = name;
        }
        if (partName == null) {
            throw new IllegalArgumentException("partName null");
        } else {
            this.partName = partName;
        }
        if (schemaFragment == null) {
            throw new IllegalArgumentException("schemaFragment null");
        } else {
            this.schemaFragment = schemaFragment;
        }
    }

    /**
     * Gets message fragment name. Name of <code>SchemaProperty</code> defined on referenced
     * schema fragment serves a proxy for message fragment name. Name is unique within context
     * of parent message.
     * 
     * @return qualified name
     */
    public final QName getName() {
        return name;
    }

    /**
     * Identifies name of abstract message part which contains fragment. The context
     * <code>Node</code> from which fragment expression is evaluated is a document
     * information item which contains a single element information item representing root
     * element of message part.
     * 
     * @return String partName.
     */
    public final String getPartName() {
        return partName;
    }

    /**
     * Gets referenced <code>SchemaFragment</code> which statically describes this message
     * fragment.
     * 
     * @return schemaFragment
     */
    public SchemaFragment getSchemaFragment() {
        return schemaFragment;
    }

    public final FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(rootName.getLocalPart());
            pointerPart.append("/");
            pointerPart.append("xmlns(ns=");
            pointerPart.append(name.getNamespaceURI());
            pointerPart.append(")");
            pointerPart.append("ns:" + name.getLocalPart());
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA,
                    "interfaceMessageFragment", pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(InterfaceMessage parent) {
        this.parent = parent;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(name);
        sb.append(System.getProperty("line.separator"));
        sb.append("partName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(partName);
        return sb.toString();
    }

    @Override
    public final int hashCode() {
        return hashCode(name) + hashCode(partName);
    }

    @Override
    public final boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof MessageFragment)) {
            return false;
        }

        if ((name == ((MessageFragment)obj).name || (name != null && name.equals(((MessageFragment)obj).name)))
                && (partName == ((MessageFragment)obj).partName || (partName != null && partName
                        .equals(((MessageFragment)obj).partName)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}