/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.policy.rt;

import java.util.Collection;
import java.util.Set;

import org.bluestemsoftware.specification.eoa.ext.policy.PolicyFactory;
import org.w3c.dom.Element;

/**
 * Models a ws-policy expression, i.e. an expression with a potentially empty set of policy
 * alternatives. Policy is used to enable/constrain <code>FeatureModule</code> components.
 * The rules which define semantics of a <code>PolicyExpression</code> are extensible, i.e.
 * in accordance with current and future versions of the ws-policy specification.
 * <p>
 * Note that policy expressions are immutable, i.e. policy operations MUST NOT modify the
 * operands, i.e. the expressions used as inputs to the operation and MUST return a new
 * instance of <code>PolicyExpression</code>.
 * 
 * @see PolicyFactory
 */
public interface PolicyExpression extends Element {

    /**
     * Returns ref URI which identifies expression type, i.e. language.
     * 
     * @return
     */
    public String getType();

    /**
     * Gets expression 'id'.
     * 
     * @return id expression 'id' or <code>null</code> if undefined.
     */

    public String getExpressionID();

    /**
     * Returns all <code>PolicyAlternatives</code> defined by this expression.
     * 
     * @return Collection of <code>PolicyAlternatives</code> or an empty Collection if no
     *         choices defined. Note that an empty Collection implies no admissable behavior.
     */
    public Collection<PolicyAlternative> getAlternatives();

    /**
     * Indicates whether or not policy expression is supported.
     * 
     * @return true if at least one alternative is supported.
     * 
     * @see PolicyAlternative#isSupported()
     */
    public boolean isSupported();

    /**
     * Returns the normal form of this expression as a new <code>PolicyExpression</code>
     * instance.
     * <p>
     * The rules used to normalize expression are specified within ws-policy specification,
     * which is an extension to this specification as identified by {@link #getType()}.
     * 
     * 
     * @return A new <code>PolicyExpression</code> which is result of normalization operation
     */
    public PolicyExpression normalizeExpression();

    /**
     * Intersects a <code>PolicyExpression</code> with this <code>PolicyExpression</code>
     * and returns the result. This <code>PolicyExpression</code> instance and the
     * <code>PolicyExpression</code> expression with which intersection is performed MUST NOT
     * be modified as a result of the intersection.
     * <p>
     * The rules used to construct result are specified within ws-policy specification, which
     * is an extension to this specification as identified by {@link #getType()}.
     * 
     * @param policyExpression
     *        The <code>PolicyExpression</code> with which to intersect, which MUST be same
     *        type as this <code>PolicyExpression</code> as indicated by {@link #getType()}.
     * 
     * @return A new <code>PolicyExpression</code> which is result of intersection operation
     */
    public PolicyExpression intersectExpression(PolicyExpression policyExpression);

    /**
     * Merges a <code>PolicyExpression</code> with this <code>PolicyExpression</code>.
     * This <code>PolicyExpression</code> instance and the <code>PolicyExpression</code>
     * with which merge is performed MUST NOT be modified as a result of the merge operation.
     * <p>
     * The rules used to construct result are specified within ws-policy specification, which
     * is an extension to this specification as identified by {@link #getType()}.
     * 
     * @param policyExpression
     *        The <code>PolicyExpression</code> to merge, which MUST be same type as this
     *        <code>PolicyExpression</code> as indicated by {@link #getType()}.
     * @return A new <code>PolicyExpression</code> which is result of merge operation
     * 
     */
    public PolicyExpression mergeExpression(PolicyExpression policyExpression);

    /**
     * Filters, i.e. removes, assertions from alternatives contained within normal form of this
     * expression other than those with a namespace matching a namespace defined within
     * indicated collection. Can be used to 'reverse' a merge operation, e.g. when an
     * <code>EffectivePolicy</code> expression contains assertions from vocabularies not
     * understood by a specific <code>Feature</code>.
     * <p>
     * Note that if, as a result of the filter operation, duplicate alternatives exist, they
     * MUST be removed.
     * <p>
     * Note that if, as a result of the filter operation, an empty alternative remains, it MUST
     * be preserved.
     * <p>
     * Note that this instance MUST NOT be modified as a result of the filter operation.
     * <p>
     * Note that assertions defined within nested policy expression(s), if any, MUST NOT be
     * filtered.
     * <p>
     * Note that ID attribute is undefined for result.
     * 
     * @param policyVocabularyNamespaces
     * @return A new <code>PolicyExpression</code> which is result of filter operation or
     *         <code>null</code> if result contains no matching assertions.
     */
    public PolicyExpression filterExpression(Set<String> policyVocabularyNamespaces);
    

    /**
     * Creates a copy of this expression. 
     * 
     * @return a deep copy of expression. Note that the returned expression MUST NOT be
     * normalized, i.e. it MUST be an exact copy of this expression.
     */
    public PolicyExpression copy();
 
}
