/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.types;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.ComponentReference.SchemaPropertyReference;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;

/**
 * Models a schema <code>Fragment</code>, i.e. an embedded instance of
 * {@link SchemaProperty} which MAY be extracted from runtime instance of parent
 * <code>SchemaComponent</code>, e.g. an element information item, using indicated
 * <code>FragmentExpression</code>.
 * <p>
 * The following restrictions MUST be enforced by a compliant implementation:
 * <p>
 * Recursive references MUST be detected, i.e. referenced <code>Property</code> MUST NOT
 * reference parent <code>SchemaComponent</code>.
 * <p>
 * Name of referenced <code>Property</code> MUST be unique within context of parent
 * <code>SchemaComponent</code>.
 * <p>
 * A <code>SchemaFragment</code> MUST be 'schema valid' within context of parent
 * <code>SchemaComponent</code>.
 */

public final class SchemaFragment extends NestedComponent<Schema, SchemaComponent> implements SchemaPropertyReference {

    private String propertyNamespace;
    private String propertyName;
    private SchemaProperty property;
    private ComponentName schemaName;
    private Schema schema;
    private FragmentExpression fragmentExpression;

    public SchemaFragment(String propertyNamespace, String propertyName, FragmentExpression fragmentExpression) {
        if (propertyNamespace == null) {
            throw new IllegalArgumentException("propertyNamespace null");
        } else {
            this.propertyNamespace = propertyNamespace;
        }
        if (propertyName == null) {
            throw new IllegalArgumentException("propertyName null");
        } else {
            this.propertyName = propertyName;
        }
        if (fragmentExpression == null) {
            throw new IllegalArgumentException("fragmentExpression null");
        } else {
            this.fragmentExpression = fragmentExpression;
        }
        // note that local name of referenced schema is unknown until
        // reference is established, at which point it will be updated
        schemaName = new ComponentName(propertyNamespace, "foo");
    }

    /**
     * Gets referenced <code>Property</code>.
     * 
     * @return Property
     */
    public SchemaProperty getProperty() {
        return property;
    }

    /**
     * Gets propertyNamespace within which referenced property is defined.
     * @return
     */
    public String getSchemaPropertyNamespace() {
        return propertyNamespace;
    }

    /**
     * Returns expression used to extract runtime property instance.
     * 
     * @return
     */
    public FragmentExpression getFragmentExpression() {
        return fragmentExpression;
    }
    
    /**
     * Gets qualified name of referenced schema property.
     * 
     * @return
     */
    public QName getSchemaPropertyQName() {
        return new QName(propertyNamespace, propertyName);
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference.SchemaPropertyReference#getSchemaPropertyName()
     */
    public String getSchemaPropertyName() {
        return propertyName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponent()
     */
    public Schema getReferencedComponent() {
        return schema;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentName()
     */
    public ComponentName getReferencedComponentName() {
        return schemaName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentType()
     */
    public ComponentType getReferencedComponentType() {
        return ComponentType.SCHEMA;
    }

    void establishReference() throws DeploymentException {
        
        if (property != null) {
            return;
        }

        // see if we're referencing a property defined within parent
        // schema. otherwise a circular reference exception is thrown

        Schema temp = (Schema)getRootComponent();
        if (temp.getName().getNamespaceURI().equals(propertyNamespace)) {
            property = temp.getProperty(propertyName);
        }

        if (property != null) {
            return;
        }

        schema = getRootComponent().getComponentContext().findReferencedComponent(this);
        if (schema == null) {
            throw new DeploymentException("ComponentReference '"
                    + getFragmentIdentifier()
                    + "' is invalid. Schema which defines property "
                    + getSchemaPropertyQName()
                    + " is undefined.");
        } else {
            property = schema.getProperty(propertyName);
            // now that localname of schema is defined update our name
            schemaName = schema.getName();
        }

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentClass()
     */
    public Class<Schema> getReferencedComponentClass() {
        return Schema.class;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     */
    public final FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(parent.getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append("xmlns(tns=");
            pointerPart.append(propertyNamespace);
            pointerPart.append(")");
            pointerPart.append("tns:");
            pointerPart.append(propertyName);
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA, "schemaFragment",
                    pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(SchemaComponent parent) {
        this.parent = parent;
    }

    void validate() throws DeploymentException {
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("propertyName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(new QName(propertyNamespace, propertyName));
        return sb.toString();
    }

    @Override
    public final int hashCode() {
        return hashCode(propertyName);
    }

    @Override
    public final boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof SchemaFragment)) {
            return false;
        }

        if ((propertyName == ((SchemaFragment)obj).propertyName || (propertyName != null && propertyName
                .equals(((SchemaFragment)obj).propertyName)))

        ) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}