/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.types;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.ComponentReference.SchemaComponentReference;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.types.SchemaComponent.Type;

/**
 * Models a <code>Property</code> component which is defined via a <code>Schema</code>
 * extension. A <code>Property</code> 'aliases' a {@link SchemaComponent} by providing a
 * unique, qualified name which SHOULD be semantically significant from name of referenced
 * <code>SchemaComponent</code>.
 * <p>
 * As an example, the simple <code>TypeDefinition</code> xs:string carries no semantic
 * significance, whereas the hypothetical property ns01:purchaseOrderID, which is derived from
 * xs:string, does.
 */
public final class SchemaProperty extends NestedComponent<Schema, Schema> implements SchemaComponentReference {

    private String propertyName;
    private QName propertyQName;
    private String schemaComponentNamespace;
    private Type schemaComponentType;
    private String schemaComponentName;
    private ComponentName schemaName;
    private Schema schema;
    private SchemaComponent schemaComponent;

    public SchemaProperty(String propertyName, String schemaComponentNamespace, Type schemaComponentType,
            String schemaComponentName) {
        super();
        if (propertyName == null) {
            throw new IllegalArgumentException("name null");
        } else {
            this.propertyName = propertyName;
        }
        if (schemaComponentNamespace == null) {
            throw new IllegalArgumentException("schemaComponentNamespace null");
        } else {
            this.schemaComponentNamespace = schemaComponentNamespace;
        }
        if (schemaComponentType == null) {
            throw new IllegalArgumentException("schemaComponentType null");
        } else {
            this.schemaComponentType = schemaComponentType;
        }
        if (schemaComponentName == null) {
            throw new IllegalArgumentException("schemaComponentName null");
        } else {
            this.schemaComponentName = schemaComponentName;
        }
        // note that local name of referenced schema is unknown until
        // reference is established, at which point it will be updated
        schemaName = new ComponentName(schemaComponentNamespace, "foo");
    }

    /**
     * Gets property name which MUST be unique within context of parent <code>Schema</code>.
     * 
     * @return name
     */
    public String getName() {
        return propertyName;
    }
    
    /**
     * Convenience method which retrieves qualified name of <code>Property</code>.
     * @return
     */
    public QName getQName() {
        return propertyQName;
    }
    
    /**
     * Gets target namespace of referenced schema component.
     * @return
     */
    public String getSchemaComponentNamespace() {
        return schemaComponentNamespace;
    }
    
    /**
     * Convenience method which retrieves qualified name of referenced schema component.
     * @return
     */
    public QName getSchemaComponentQName() {
        return new QName(schemaComponentNamespace, schemaComponentName);
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference.SchemaComponentReference#getSchemaComponentName()
     */
    public String getSchemaComponentName() {
        return schemaComponentName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference.SchemaComponentReference#getSchemaComponentType()
     */
    public Type getSchemaComponentType() {
        return schemaComponentType;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentClass()
     */
    public Class<Schema> getReferencedComponentClass() {
        return Schema.class;
    }

    /**
     * Gets referenced <code>SchemaComponent</code>.
     * 
     * @return SchemaComponent referenced <code>SchemaComponent</code>
     */
    public SchemaComponent getSchemaComponent() {
        return schemaComponent;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponent()
     */
    public Schema getReferencedComponent() {
        return schema;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentName()
     */
    public ComponentName getReferencedComponentName() {
        return schemaName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentType()
     */
    public ComponentType getReferencedComponentType() {
        return ComponentType.SCHEMA;
    }

    void establishReference() throws DeploymentException {
        
        if (schemaComponent != null) {
            return;
        }
        
        // see if we're referencing a component defined within parent
        // schema. otherwise a circular reference exception is thrown
        
        Schema temp = (Schema)getRootComponent();
        if (temp.getName().getNamespaceURI().equals(schemaComponentNamespace)) {
            schemaComponent = temp.getSchemaComponent(schemaComponentType, schemaComponentName);
        }
         
        if (schemaComponent != null) {
            return;
        }
        
        schema = temp.getComponentContext().findReferencedComponent(this);
        if (schema == null) {
            throw new DeploymentException("ComponentReference '"
                    + getFragmentIdentifier()
                    + "' is invalid. Schema which defines schema component "
                    + new QName(schemaComponentNamespace, schemaComponentName)
                    + " typed as a "
                    + schemaComponentType
                    + ", is undefined.");
        } else {
            schemaComponent = schema.getSchemaComponent(schemaComponentType, schemaComponentName);
            // now that localname of schema is defined update our name
            schemaName = schema.getName();
        }
        
    }

    public final FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(propertyName);
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA, "schemaProperty",
                    pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(Schema parent) {
        this.parent = parent;
        this.propertyQName = new QName(parent.getName().getNamespaceURI(), propertyName);
    }

    void validate() throws DeploymentException {
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(propertyName);
        sb.append(System.getProperty("line.separator"));
        sb.append("schemaComponentName:");
        sb.append(System.getProperty("line.separator"));
        sb.append(schemaComponentName);
        sb.append(System.getProperty("line.separator"));
        sb.append("schemaComponentType:");
        sb.append(System.getProperty("line.separator"));
        sb.append(schemaComponentType);
        return sb.toString();
    }

    @Override
    public final int hashCode() {
        return hashCode(propertyName) + hashCode(schemaComponentName) + hashCode(schemaComponentType);
    }

    @Override
    public final boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof SchemaProperty)) {
            return false;
        }

        if ((propertyName == ((SchemaProperty)obj).propertyName || (propertyName != null && propertyName.equals(((SchemaProperty)obj).propertyName)))
                && (schemaComponentName == ((SchemaProperty)obj).schemaComponentName || (schemaComponentName != null && schemaComponentName
                        .equals(((SchemaProperty)obj).schemaComponentName)))
                && (schemaComponentType == ((SchemaProperty)obj).schemaComponentType || (schemaComponentType != null && schemaComponentType
                        .equals(((SchemaProperty)obj).schemaComponentType)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}
