/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext;



/**
 * Models a 'nested' <code>Extension</code>, i.e. an <code>Extension</code> (child) which is nested within
 * another <code>Extension</code> (parent).
 */
public abstract class NestedExtension<R extends RootExtension, P extends Extension> implements Extension {

    protected R rootExtension;
    protected Provider provider;
    protected P parent;
    
    public NestedExtension(Provider provider) {
        if (provider == null) {
            throw new IllegalArgumentException("extensionProvider null");
        } else {
            this.provider = provider;
            this.provider.spi_setConsumer(this);
        }
    }

    /**
     * Retrieves parent extension, i.e. the extension within which this extension is 'nested.'
     * 
     * @return The parent extension.
     */
    public final P getParent() {
        return parent;
    }

    /**
     * Convenience method which returns top-level extension within which this extension is
     * nested.
     * <p>
     * Note that if extension is nested directly beneath a top-level extension, then this
     * method returns same result as {@link #getParent()}.
     * 
     * @return RootExtension
     * @throws IllegalStateException
     *         If <code>RootExtension</code> is undefined.
     */
    @SuppressWarnings("unchecked")
    public final R getRootExtension() {
        if (rootExtension == null) {
            if (parent == null) {
                throw new IllegalStateException("Parent extension not yet set.");
            }
            Extension temp = parent;
            while (temp instanceof NestedExtension) {
                temp = ((NestedExtension)temp).getParent();
            }
            if (temp != null) {
                rootExtension = (R)temp;
            } else {
                throw new IllegalStateException("RootExtension is undefined.");
            }
        }
        return rootExtension;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.Extension#getExtensionFactory()
     */
    public final ExtensionFactory getExtensionFactory() {
        return getRootExtension().getExtensionFactory();
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.Extension#getExtensionProvider()
     */
    public final Provider getExtensionProvider() {
        return provider;
    }  
   
 }