/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.connector;

import java.util.Properties;

import org.bluestemsoftware.specification.eoa.ext.ExtensionFactory;
import org.bluestemsoftware.specification.eoa.system.container.ConnectionSupport;
import org.bluestemsoftware.specification.eoa.system.container.Connector;
import org.bluestemsoftware.specification.eoa.system.container.Container;

/**
 * A factory for type specific instances of {@link Connector}. Used by Container to
 * instantiate <code>Connector</code> instances retrievable via
 * {@link Container#getConnector(String)}.
 */
public abstract class ConnectorFactory extends ExtensionFactory {

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends ExtensionFactory.Provider {

        public String spi_getConnectorImpl();

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ExtensionFactory#getType()
     */
    public final Type getType() {
        return Type.CONNECTOR;
    }

    /**
     * Returns a URI which uniquely identifies implementation. Value MUST be unique within
     * context of a specfic <code>ConnectorFactory</code> as identified by
     * {@link #getExtensionType()). Value SHOULD be globally unique.
     * <p>
     * More than one implementation of a <code>ConnectorFactory</code> may exist within
     * context of <code>Container</code>, e.g. assuming <code>ConnectorFactory</code> is a
     * factory for <code>javax.sql.DataSource</code> instances, an impl for each supported
     * DBMS vendor MAY be loaded.
     * 
     * @return A URI formatted as a String.
     */
    public String getConnectorImpl() {
        if (factoryContext == null) {
            throw new IllegalStateException("Factory has no context.");
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factoryContext.getClassLoader());
            return ((Provider)provider).spi_getConnectorImpl();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Creates a type specific instance of {@link Connector}. Used by Container to instantiate
     * <code>Connector</code> instances retrievable via
     * {@link Container#getConnector(String)}.
     * 
     * @param name
     *        <code>Connector</code> name which MUST be unique within context of
     *        <code>Container</code>. Value MUST be set by factory on returned
     *        <code>Connector</code>. Value SHOULD be immutable, i.e. set using a
     *        constructor argument.
     * @param connectionSupport
     *        Supplied by <code>Container</code>. Provides 'hooks' used by
     *        <code>Connector</code> instances to access QOS related features, e.g. pooling
     *        support, transaction support, security, etc ...
     * @param Properties
     *        implementation specific properties used to initialize <code>Connector</code>
     *        instance or null if no properties defined, i.e. either none are required or
     *        defaults are appropriate
     * @param isXA
     *        indicates whether or not returned <code>Connector</code> instance is XA
     *        capable, i.e. generates connections which can participate in a distributed
     *        transaction. If false, <code>Connector</code> supports local transactions.
     * @return Connector an EOA <code>Connector</code> component.
     */
    public abstract Connector createConnector(String name, ConnectionSupport connectionSupport, Properties properties, boolean isXA) throws ConnectorException;
    
}
