/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.system.server;

import java.util.Set;

import org.bluestemsoftware.specification.eoa.component.engine.rt.BindingModule;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.FeatureModule;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.SystemFault;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ActionPolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.UnsupportedPolicyException;
import org.bluestemsoftware.specification.eoa.ext.feature.FeatureException;
import org.bluestemsoftware.specification.eoa.ext.feature.FeatureFactory;

/**
 * Identifies a specific type of <code>Feature</code>, i.e. a feature whose behavior is
 * implemented on a per message basis within 'binding layer' via one or more
 * {@link BindingModule} instances.
 */
public abstract class BindingFeature extends Feature implements ModulableFeature {

    protected Set<String> policyVocabularyNamespaces;
    protected String bindingType;
    protected Boolean isNative;

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends Feature.Provider {

        public boolean spi_isNative();

        public Set<String> spi_getPolicyVocabularyNamespaces();

        public void spi_destroyFeatureModule(FeatureModule featureModule);

    }

    public BindingFeature(FeatureFactory factory, Provider provider, String bindingType) {
        super(factory, provider);
        if (bindingType == null) {
            throw new IllegalArgumentException("bindingType null");
        } else {
            this.bindingType = bindingType;
        }
    }

    /**
     * Indicates whether or not feature is natively implemented by <code>Binding</code>.
     * 
     * 
     * @return if true feature MUST be natively implemented by <code>Binding</code>. if
     *         false feature MUST be implemented by <code>BindingFeature</code>.
     */
    public boolean isNative() {
        if (provider == null) {
            throw new IllegalStateException("provider not set.");
        }
        if (isNative == null) {
            Thread thread = Thread.currentThread();
            ClassLoader cl = thread.getContextClassLoader();
            try {
                thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
                isNative = ((Provider)provider).spi_isNative();
            } finally {
                thread.setContextClassLoader(cl);
            }
        }
        return isNative;
    }

    /**
     * Returns type ref URI which identifies <code>Binding</code> type 'understood' by
     * binding modules created by feature.
     * 
     * @return
     */
    public String getBindingType() {
        return bindingType;
    }

    /**
     * Processes domain specific policy which enables/constrains <code>Feature</code> and
     * creates one or more instances of <code>BindingModule</code> which implement
     * <code>Feature</code> on a per message basis by processing message instances associated
     * with action identified by {@link EndpointActionReference#getAction()}.
     * 
     * @param endpointActionReference
     *        targeted component upon which modules will be installed. References action to
     *        which 'private' and 'public' policy applies. Note that the returned module MUST
     *        be compatible with <code>Binding</code> employed by referenced endpoint.
     * 
     * @param privatePolicy
     *        Aggregates 'private' effective policy expressions from all four policy subjects
     *        which MAY affect processing of messages associated with referenced action. The
     *        contained expressions have been filtered such that they contain only assertions
     *        understood by <code>ModulableFeature</code>, i.e. assertions which match a
     *        namespace returned by {@link #getPolicyVocabularyNamespaces()}.
     *        <p>
     *        Note that 'private' policy MAY be used to negotiate alternatives defined on
     *        referenced action, i.e. referenced action may advertise that it supports more
     *        than one capability, in which case 'private' policy would be intersected with
     *        policy defined on referenced action to 'negotiate' a preferred alternative.
     * 
     * @param publicPolicy
     *        Aggregates 'public' effective policy expressions from all four policy subjects
     *        which MAY affect processing of messages associated with referenced action. The
     *        contained expressions have been filtered such that they contain only assertions
     *        understood by <code>ModulableFeature</code>, i.e. assertions which match a
     *        namespace returned by {@link #getPolicyVocabularyNamespaces()}.
     * 
     * @return BindingModule instance
     * @throws FeatureException
     *         if an error occurs creating module(s).
     * @throws UnsupportedPolicyException
     *         Which MUST be thrown by Feature if it is unable to comply with policy.
     */
    public abstract BindingModule createBindingModule(EndpointActionReference endpointActionReference, ActionPolicy privatePolicy, ActionPolicy publicPolicy) throws FeatureException, UnsupportedPolicyException;

    /**
     * Returns <code>BindingModule</code> configured using a default configuration which is
     * used to process undeclared action {@link SystemFault#ACTION}.
     * 
     * @return default binding module
     * @throws FeatureException
     *         if an error occurs creating module
     * 
     * @see EndpointReference#getDefaultBindingModules()
     */
    public abstract BindingModule createDefaultBindingModule() throws FeatureException;

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.system.server.ModulableFeature#getPolicyVocabularyNamespaces()
     */
    public Set<String> getPolicyVocabularyNamespaces() {
        if (provider == null) {
            throw new IllegalStateException("provider not set.");
        }
        if (policyVocabularyNamespaces == null) {
            Thread thread = Thread.currentThread();
            ClassLoader cl = thread.getContextClassLoader();
            try {
                thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
                policyVocabularyNamespaces = ((Provider)provider).spi_getPolicyVocabularyNamespaces();
            } finally {
                thread.setContextClassLoader(cl);
            }
        }
        return policyVocabularyNamespaces;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.system.server.ModulableFeature#destroyFeatureModule(org.bluestemsoftware.specification.eoa.component.engine.rt.FeatureModule)
     */
    public void destroyFeatureModule(FeatureModule featureModule) {
        if (provider == null) {
            throw new IllegalStateException("provider not set.");
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_destroyFeatureModule(featureModule);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

}
