/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.system.server;

import org.w3c.dom.Element;

/**
 * Identifies a feature which MUST be enabled on server, i.e. feature MUST be explicitly
 * defined within server definition for it to be enabled.
 */
public final class EnabledFeature {

    private String type;
    private String impl;
    private Element configuration;

    /**
     * Constructs a <code>EnabledFeature</code> object.
     * 
     * @param type
     *        Identifies <code>Feature</code> type.
     * @param impl
     *        Identifies a specific implementation of <code>Feature</code> type or null if
     *        implementation chosen is arbitrary.
     * @param configuration
     *        Root element of an impl specific configuration document used by
     *        <code>Server</code> to initialize managed <code>Feature</code> or
     *        <code>null</code> if <code>Feature</code> requires no configuration or if
     *        default configuration is acceptable.
     */
    public EnabledFeature(String type, String impl, Element configuration) {
        if (type == null) {
            throw new IllegalArgumentException("type null");
        } else {
            this.type = type;
        }
        this.impl = impl;
        this.configuration = configuration;
    }

    /**
     * Root element of an impl specific configuration document used by <code>Server</code> to
     * initialize managed <code>Feature</code>.
     * 
     * @return Element or <code>null</code> if <code>Feature</code> requires no
     *         configuration or if default configuration is acceptable.
     */
    public Element getConfiguration() {
        return configuration;
    }

    /**
     * Identifies a <code>Feature</code> provider.
     * 
     * @return URI or null if no provider specified by user
     */
    public String getFeatureImpl() {
        return impl;
    }

    /**
     * Identifies <code>Feature</code> type.
     * 
     * @return URI
     */
    public String getType() {
        return type;
    }

}
